import sys

import numpy as np

import openmdao.api as om

import pycycle.api as pyc


class HBTF(pyc.Cycle):

    def initialize(self):
        # Initialize the model here by setting option variables such as a switch for design vs off-des cases
        self.options.declare('throttle_mode', default='T4', values=['T4', 'percent_thrust'])

        super().initialize()


    def setup(self):
        #Setup the problem by including all the relavant components here - comp, burner, turbine etc

        #Create any relavent short hands here:
        design = self.options['design']

        USE_TABULAR = False
        if USE_TABULAR:
            self.options['thermo_method'] = 'TABULAR'
            self.options['thermo_data'] = pyc.AIR_JETA_TAB_SPEC
            FUEL_TYPE = 'FAR'
        else:
            self.options['thermo_method'] = 'CEA'
            self.options['thermo_data'] = pyc.species_data.janaf
            FUEL_TYPE = 'Jet-A(g)'


        #Add subsystems to build the engine deck:
        self.add_subsystem('fc', pyc.FlightConditions())
        self.add_subsystem('inlet', pyc.Inlet())

        # Note variable promotion for the fan --
        # the LP spool speed and the fan speed are INPUTS that are promoted:
        # Note here that promotion aliases are used. Here Nmech is being aliased to LP_Nmech
        # check out: https://openmdao.org/newdocs/versions/latest/features/core_features/working_with_groups/add_subsystem.html?highlight=alias
        self.add_subsystem('fan', pyc.Compressor(map_data=pyc.FanMap,
                                        bleed_names=[], map_extrap=True), promotes_inputs=[('Nmech','LP_Nmech')])
        self.add_subsystem('splitter', pyc.Splitter())
        self.add_subsystem('duct4', pyc.Duct())
        self.add_subsystem('lpc', pyc.Compressor(map_data=pyc.LPCMap,
                                        map_extrap=True),promotes_inputs=[('Nmech','LP_Nmech')])
        self.add_subsystem('duct6', pyc.Duct())
        self.add_subsystem('hpc', pyc.Compressor(map_data=pyc.HPCMap,
                                        bleed_names=['cool1','cool2','cust'], map_extrap=True),promotes_inputs=[('Nmech','HP_Nmech')])
        self.add_subsystem('bld3', pyc.BleedOut(bleed_names=['cool3','cool4']))
        self.add_subsystem('burner', pyc.Combustor(fuel_type=FUEL_TYPE))
        self.add_subsystem('hpt', pyc.Turbine(map_data=pyc.HPTMap,
                                        bleed_names=['cool3','cool4'], map_extrap=True),promotes_inputs=[('Nmech','HP_Nmech')])
        self.add_subsystem('duct11', pyc.Duct())
        self.add_subsystem('lpt', pyc.Turbine(map_data=pyc.LPTMap,
                                        bleed_names=['cool1','cool2'], map_extrap=True),promotes_inputs=[('Nmech','LP_Nmech')])
        self.add_subsystem('duct13', pyc.Duct())
        self.add_subsystem('core_nozz', pyc.Nozzle(nozzType='CV', lossCoef='Cv'))

        self.add_subsystem('byp_bld', pyc.BleedOut(bleed_names=['bypBld']))
        self.add_subsystem('duct15', pyc.Duct())
        self.add_subsystem('byp_nozz', pyc.Nozzle(nozzType='CV', lossCoef='Cv'))

        #Create shaft instances. Note that LP shaft has 3 ports! => no gearbox
        self.add_subsystem('lp_shaft', pyc.Shaft(num_ports=3),promotes_inputs=[('Nmech','LP_Nmech')])
        self.add_subsystem('hp_shaft', pyc.Shaft(num_ports=2),promotes_inputs=[('Nmech','HP_Nmech')])
        self.add_subsystem('perf', pyc.Performance(num_nozzles=2, num_burners=1))

        # Now use the explicit connect method to make connections -- connect(<from>, <to>)

        #Connect the inputs to perf group
        self.connect('inlet.Fl_O:tot:P', 'perf.Pt2')
        self.connect('hpc.Fl_O:tot:P', 'perf.Pt3')
        self.connect('burner.Wfuel', 'perf.Wfuel_0')
        self.connect('inlet.F_ram', 'perf.ram_drag')
        self.connect('core_nozz.Fg', 'perf.Fg_0')
        self.connect('byp_nozz.Fg', 'perf.Fg_1')

        #LP-shaft connections
        self.connect('fan.trq', 'lp_shaft.trq_0')
        self.connect('lpc.trq', 'lp_shaft.trq_1')
        self.connect('lpt.trq', 'lp_shaft.trq_2')
        #HP-shaft connections
        self.connect('hpc.trq', 'hp_shaft.trq_0')
        self.connect('hpt.trq', 'hp_shaft.trq_1')
        #Ideally expanding flow by conneting flight condition static pressure to nozzle exhaust pressure
        self.connect('fc.Fl_O:stat:P', 'core_nozz.Ps_exhaust')
        self.connect('fc.Fl_O:stat:P', 'byp_nozz.Ps_exhaust')

        #Create a balance component
        # Balances can be a bit confusing, here's some explanation -
        #   State Variables:
        #           (W)        Inlet mass flow rate to implictly balance thrust
        #                      LHS: perf.Fn  == RHS: Thrust requirement (set when TF is instantiated)
        #
        #           (FAR)      Fuel-air ratio to balance Tt4
        #                      LHS: burner.Fl_O:tot:T  == RHS: Tt4 target (set when TF is instantiated)
        #
        #           (lpt_PR)   LPT press ratio to balance shaft power on the low spool
        #           (hpt_PR)   HPT press ratio to balance shaft power on the high spool
        # Ref: look at the XDSM diagrams in the pyCycle paper and this:
        # http://openmdao.org/newdocs/versions/latest/features/building_blocks/components/balance_comp.html

        balance = self.add_subsystem('balance', om.BalanceComp())
        if design:
            balance.add_balance('W', units='lbm/s', eq_units='lbf')
            #Here balance.W is implicit state variable that is the OUTPUT of balance object
            self.connect('balance.W', 'fc.W') #Connect the output of balance to the relevant input
            self.connect('perf.Fn', 'balance.lhs:W')       #This statement makes perf.Fn the LHS of the balance eqn.
            self.promotes('balance', inputs=[('rhs:W', 'Fn_DES')])

            balance.add_balance('FAR', eq_units='degR', lower=1e-4, val=.017)
            self.connect('balance.FAR', 'burner.Fl_I:FAR')
            self.connect('burner.Fl_O:tot:T', 'balance.lhs:FAR')
            self.promotes('balance', inputs=[('rhs:FAR', 'T4_MAX')])

            # Note that for the following two balances the mult val is set to -1 so that the NET torque is zero
            balance.add_balance('lpt_PR', val=1.5, lower=1.001, upper=8,
                                eq_units='hp', use_mult=True, mult_val=-1)
            self.connect('balance.lpt_PR', 'lpt.PR')
            self.connect('lp_shaft.pwr_in_real', 'balance.lhs:lpt_PR')
            self.connect('lp_shaft.pwr_out_real', 'balance.rhs:lpt_PR')

            balance.add_balance('hpt_PR', val=1.5, lower=1.001, upper=8,
                                eq_units='hp', use_mult=True, mult_val=-1)
            self.connect('balance.hpt_PR', 'hpt.PR')
            self.connect('hp_shaft.pwr_in_real', 'balance.lhs:hpt_PR')
            self.connect('hp_shaft.pwr_out_real', 'balance.rhs:hpt_PR')

        else:

            #In OFF-DESIGN mode we need to redefine the balances:
            #   State Variables:
            #           (W)        Inlet mass flow rate to balance core flow area
            #                      LHS: core_nozz.Throat:stat:area == Area from DESIGN calculation
            #
            #           (FAR)      Fuel-air ratio to balance Thrust req.
            #                      LHS: perf.Fn  == RHS: Thrust requirement (set when TF is instantiated)
            #
            #           (BPR)      Bypass ratio to balance byp. noz. area
            #                      LHS: byp_nozz.Throat:stat:area == Area from DESIGN calculation
            #
            #           (lp_Nmech)   LP spool speed to balance shaft power on the low spool
            #           (hp_Nmech)   HP spool speed to balance shaft power on the high spool

            if self.options['throttle_mode'] == 'T4':
                balance.add_balance('FAR', val=0.017, lower=1e-4, eq_units='degR')
                self.connect('balance.FAR', 'burner.Fl_I:FAR')
                self.connect('burner.Fl_O:tot:T', 'balance.lhs:FAR')
                self.promotes('balance', inputs=[('rhs:FAR', 'T4_MAX')])

            elif self.options['throttle_mode'] == 'percent_thrust':
                balance.add_balance('FAR', val=0.017, lower=1e-4, eq_units='lbf', use_mult=True)
                self.connect('balance.FAR', 'burner.Fl_I:FAR')
                self.connect('perf.Fn', 'balance.rhs:FAR')
                self.promotes('balance', inputs=[('mult:FAR', 'PC'), ('lhs:FAR', 'Fn_max')])


            balance.add_balance('W', units='lbm/s', lower=10., upper=1000., eq_units='inch**2')
            self.connect('balance.W', 'fc.W')
            self.connect('core_nozz.Throat:stat:area', 'balance.lhs:W')

            balance.add_balance('BPR', lower=2., upper=10., eq_units='inch**2')
            self.connect('balance.BPR', 'splitter.BPR')
            self.connect('byp_nozz.Throat:stat:area', 'balance.lhs:BPR')

            # Again for the following two balances the mult val is set to -1 so that the NET torque is zero
            balance.add_balance('lp_Nmech', val=1.5, units='rpm', lower=500., eq_units='hp', use_mult=True, mult_val=-1)
            self.connect('balance.lp_Nmech', 'LP_Nmech')
            self.connect('lp_shaft.pwr_in_real', 'balance.lhs:lp_Nmech')
            self.connect('lp_shaft.pwr_out_real', 'balance.rhs:lp_Nmech')

            balance.add_balance('hp_Nmech', val=1.5, units='rpm', lower=500., eq_units='hp', use_mult=True, mult_val=-1)
            self.connect('balance.hp_Nmech', 'HP_Nmech')
            self.connect('hp_shaft.pwr_in_real', 'balance.lhs:hp_Nmech')
            self.connect('hp_shaft.pwr_out_real', 'balance.rhs:hp_Nmech')

            # Specify the order in which the subsystems are executed:

            # self.set_order(['balance', 'fc', 'inlet', 'fan', 'splitter', 'duct4', 'lpc', 'duct6', 'hpc', 'bld3', 'burner', 'hpt', 'duct11',
            #                 'lpt', 'duct13', 'core_nozz', 'byp_bld', 'duct15', 'byp_nozz', 'lp_shaft', 'hp_shaft', 'perf'])

        # Set up all the flow connections:
        self.pyc_connect_flow('fc.Fl_O', 'inlet.Fl_I')
        self.pyc_connect_flow('inlet.Fl_O', 'fan.Fl_I')
        self.pyc_connect_flow('fan.Fl_O', 'splitter.Fl_I')
        self.pyc_connect_flow('splitter.Fl_O1', 'duct4.Fl_I')
        self.pyc_connect_flow('duct4.Fl_O', 'lpc.Fl_I')
        self.pyc_connect_flow('lpc.Fl_O', 'duct6.Fl_I')
        self.pyc_connect_flow('duct6.Fl_O', 'hpc.Fl_I')
        self.pyc_connect_flow('hpc.Fl_O', 'bld3.Fl_I')
        self.pyc_connect_flow('bld3.Fl_O', 'burner.Fl_I')
        self.pyc_connect_flow('burner.Fl_O', 'hpt.Fl_I')
        self.pyc_connect_flow('hpt.Fl_O', 'duct11.Fl_I')
        self.pyc_connect_flow('duct11.Fl_O', 'lpt.Fl_I')
        self.pyc_connect_flow('lpt.Fl_O', 'duct13.Fl_I')
        self.pyc_connect_flow('duct13.Fl_O','core_nozz.Fl_I')
        self.pyc_connect_flow('splitter.Fl_O2', 'byp_bld.Fl_I')
        self.pyc_connect_flow('byp_bld.Fl_O', 'duct15.Fl_I')
        self.pyc_connect_flow('duct15.Fl_O', 'byp_nozz.Fl_I')

        #Bleed flows:
        self.pyc_connect_flow('hpc.cool1', 'lpt.cool1', connect_stat=False)
        self.pyc_connect_flow('hpc.cool2', 'lpt.cool2', connect_stat=False)
        self.pyc_connect_flow('bld3.cool3', 'hpt.cool3', connect_stat=False)
        self.pyc_connect_flow('bld3.cool4', 'hpt.cool4', connect_stat=False)

        #Specify solver settings:
        newton = self.nonlinear_solver = om.NewtonSolver()
        newton.options['atol'] = 1e-8

        # set this very small, so it never activates and we rely on atol
        newton.options['rtol'] = 1e-99
        newton.options['iprint'] = 2
        newton.options['maxiter'] = 50
        newton.options['solve_subsystems'] = True
        newton.options['max_sub_solves'] = 1000
        newton.options['reraise_child_analysiserror'] = False
        # ls = newton.linesearch = BoundsEnforceLS()
        ls = newton.linesearch = om.ArmijoGoldsteinLS()
        ls.options['maxiter'] = 3
        ls.options['rho'] = 0.75
        # ls.options['print_bound_enforce'] = True

        self.linear_solver = om.DirectSolver()

        super().setup()


def viewer(prob, pt, file=sys.stdout):
    """
    print a report of all the relevant cycle properties
    """

    if pt == 'DESIGN':
        MN = prob['DESIGN.fc.Fl_O:stat:MN']
        LPT_PR = prob['DESIGN.balance.lpt_PR']
        HPT_PR = prob['DESIGN.balance.hpt_PR']
        FAR = prob['DESIGN.balance.FAR']
    else:
        MN = prob[pt+'.fc.Fl_O:stat:MN']
        LPT_PR = prob[pt+'.lpt.PR']
        HPT_PR = prob[pt+'.hpt.PR']
        FAR = prob[pt+'.balance.FAR']

    summary_data = (MN[0], prob[pt+'.fc.alt'][0], prob[pt+'.inlet.Fl_O:stat:W'][0], prob[pt+'.perf.Fn'][0],
                        prob[pt+'.perf.Fg'][0], prob[pt+'.inlet.F_ram'][0], prob[pt+'.perf.OPR'][0],
                        prob[pt+'.perf.TSFC'][0], prob[pt+'.splitter.BPR'][0])

    print(file=file, flush=True)
    print(file=file, flush=True)
    print(file=file, flush=True)
    print("----------------------------------------------------------------------------", file=file, flush=True)
    print("                              POINT:", pt, file=file, flush=True)
    print("----------------------------------------------------------------------------", file=file, flush=True)
    print("                       PERFORMANCE CHARACTERISTICS", file=file, flush=True)
    print("    Mach      Alt       W      Fn      Fg    Fram     OPR     TSFC      BPR ", file=file, flush=True)
    print(" %7.5f  %7.1f %7.3f %7.1f %7.1f %7.1f %7.3f  %7.5f  %7.3f" %summary_data, file=file, flush=True)


    fs_names = ['fc.Fl_O', 'inlet.Fl_O', 'fan.Fl_O', 'splitter.Fl_O1', 'splitter.Fl_O2',
                'duct4.Fl_O', 'lpc.Fl_O', 'duct6.Fl_O', 'hpc.Fl_O', 'bld3.Fl_O', 'burner.Fl_O',
                'hpt.Fl_O', 'duct11.Fl_O', 'lpt.Fl_O', 'duct13.Fl_O', 'core_nozz.Fl_O', 'byp_bld.Fl_O',
                'duct15.Fl_O', 'byp_nozz.Fl_O']
    fs_full_names = [f'{pt}.{fs}' for fs in fs_names]
    pyc.print_flow_station(prob, fs_full_names, file=file)

    comp_names = ['fan', 'lpc', 'hpc']
    comp_full_names = [f'{pt}.{c}' for c in comp_names]
    pyc.print_compressor(prob, comp_full_names, file=file)

    pyc.print_burner(prob, [f'{pt}.burner'], file=file)

    turb_names = ['hpt', 'lpt']
    turb_full_names = [f'{pt}.{t}' for t in turb_names]
    pyc.print_turbine(prob, turb_full_names, file=file)

    noz_names = ['core_nozz', 'byp_nozz']
    noz_full_names = [f'{pt}.{n}' for n in noz_names]
    pyc.print_nozzle(prob, noz_full_names, file=file)

    shaft_names = ['hp_shaft', 'lp_shaft']
    shaft_full_names = [f'{pt}.{s}' for s in shaft_names]
    pyc.print_shaft(prob, shaft_full_names, file=file)

    bleed_names = ['hpc', 'bld3', 'byp_bld']
    bleed_full_names = [f'{pt}.{b}' for b in bleed_names]
    pyc.print_bleed(prob, bleed_full_names, file=file)


class MPhbtf(pyc.MPCycle):

    def setup(self):

        self.pyc_add_pnt('DESIGN', HBTF(thermo_method='CEA')) # Create an instace of the High Bypass ratio Turbofan

        self.set_input_defaults('DESIGN.inlet.MN', 0.751)
        self.set_input_defaults('DESIGN.fan.MN', 0.4578)
        self.set_input_defaults('DESIGN.splitter.BPR', 5.105)
        self.set_input_defaults('DESIGN.splitter.MN1', 0.3104)
        self.set_input_defaults('DESIGN.splitter.MN2', 0.4518)
        self.set_input_defaults('DESIGN.duct4.MN', 0.3121)
        self.set_input_defaults('DESIGN.lpc.MN', 0.3059)
        self.set_input_defaults('DESIGN.duct6.MN', 0.3563)
        self.set_input_defaults('DESIGN.hpc.MN', 0.2442)
        self.set_input_defaults('DESIGN.bld3.MN', 0.3000)
        self.set_input_defaults('DESIGN.burner.MN', 0.1025)
        self.set_input_defaults('DESIGN.hpt.MN', 0.3650)
        self.set_input_defaults('DESIGN.duct11.MN', 0.3063)
        self.set_input_defaults('DESIGN.lpt.MN', 0.4127)
        self.set_input_defaults('DESIGN.duct13.MN', 0.4463)
        self.set_input_defaults('DESIGN.byp_bld.MN', 0.4489)
        self.set_input_defaults('DESIGN.duct15.MN', 0.4589)
        self.set_input_defaults('DESIGN.LP_Nmech', 4666.1, units='rpm')
        self.set_input_defaults('DESIGN.HP_Nmech', 14705.7, units='rpm')

        # --- Set up bleed values -----

        self.pyc_add_cycle_param('inlet.ram_recovery', 0.9990)
        self.pyc_add_cycle_param('duct4.dPqP', 0.0048)
        self.pyc_add_cycle_param('duct6.dPqP', 0.0101)
        self.pyc_add_cycle_param('burner.dPqP', 0.0540)
        self.pyc_add_cycle_param('duct11.dPqP', 0.0051)
        self.pyc_add_cycle_param('duct13.dPqP', 0.0107)
        self.pyc_add_cycle_param('duct15.dPqP', 0.0149)
        self.pyc_add_cycle_param('core_nozz.Cv', 0.9933)
        self.pyc_add_cycle_param('byp_bld.bypBld:frac_W', 0.005)
        self.pyc_add_cycle_param('byp_nozz.Cv', 0.9939)
        self.pyc_add_cycle_param('hpc.cool1:frac_W', 0.050708)
        self.pyc_add_cycle_param('hpc.cool1:frac_P', 0.5)
        self.pyc_add_cycle_param('hpc.cool1:frac_work', 0.5)
        self.pyc_add_cycle_param('hpc.cool2:frac_W', 0.020274)
        self.pyc_add_cycle_param('hpc.cool2:frac_P', 0.55)
        self.pyc_add_cycle_param('hpc.cool2:frac_work', 0.5)
        self.pyc_add_cycle_param('bld3.cool3:frac_W', 0.067214)
        self.pyc_add_cycle_param('bld3.cool4:frac_W', 0.101256)
        self.pyc_add_cycle_param('hpc.cust:frac_P', 0.5)
        self.pyc_add_cycle_param('hpc.cust:frac_work', 0.5)
        self.pyc_add_cycle_param('hpc.cust:frac_W', 0.0445)
        self.pyc_add_cycle_param('hpt.cool3:frac_P', 1.0)
        self.pyc_add_cycle_param('hpt.cool4:frac_P', 0.0)
        self.pyc_add_cycle_param('lpt.cool1:frac_P', 1.0)
        self.pyc_add_cycle_param('lpt.cool2:frac_P', 0.0)
        self.pyc_add_cycle_param('hp_shaft.HPX', 250.0, units='hp')

        self.od_pts = ['OD_full_pwr', 'OD_part_pwr']

        self.od_MNs = [0.8, 0.8]
        self.od_alts = [35000.0, 35000.0]
        self.od_Fn_target = [5500.0, 5300]
        self.od_dTs = [0.0, 0.0]

        self.pyc_add_pnt('OD_full_pwr', HBTF(design=False, thermo_method='CEA', throttle_mode='T4'))

        self.set_input_defaults('OD_full_pwr.fc.MN', 0.8)
        self.set_input_defaults('OD_full_pwr.fc.alt', 35000, units='ft')
        self.set_input_defaults('OD_full_pwr.fc.dTs', 0., units='degR')

        self.pyc_add_pnt('OD_part_pwr', HBTF(design=False, thermo_method='CEA', throttle_mode='percent_thrust'))

        self.set_input_defaults('OD_part_pwr.fc.MN', 0.8)
        self.set_input_defaults('OD_part_pwr.fc.alt', 35000, units='ft')
        self.set_input_defaults('OD_part_pwr.fc.dTs', 0., units='degR')

        self.connect('OD_full_pwr.perf.Fn', 'OD_part_pwr.Fn_max')

        self.pyc_use_default_des_od_conns()

        #Set up the RHS of the balances!
        self.pyc_connect_des_od('core_nozz.Throat:stat:area','balance.rhs:W')
        self.pyc_connect_des_od('byp_nozz.Throat:stat:area','balance.rhs:BPR')

        super().setup()


if __name__ == "__main__":

    import time

    prob = om.Problem()

    prob.model = mp_hbtf = MPhbtf()

    prob.setup()

    prob.set_val('DESIGN.fan.PR', 1.685)
    prob.set_val('DESIGN.fan.eff', 0.8948)

    prob.set_val('DESIGN.lpc.PR', 1.935)
    prob.set_val('DESIGN.lpc.eff', 0.9243)

    prob.set_val('DESIGN.hpc.PR', 9.369)
    prob.set_val('DESIGN.hpc.eff', 0.8707)

    prob.set_val('DESIGN.hpt.eff', 0.8888)
    prob.set_val('DESIGN.lpt.eff', 0.8996)

    prob.set_val('DESIGN.fc.alt', 35000., units='ft')
    prob.set_val('DESIGN.fc.MN', 0.8)

    prob.set_val('DESIGN.T4_MAX', 2857, units='degR')
    prob.set_val('DESIGN.Fn_DES', 5900.0, units='lbf')

    prob.set_val('OD_full_pwr.T4_MAX', 2857, units='degR')
    prob.set_val('OD_part_pwr.PC', 0.8)


    # Set initial guesses for balances
    prob['DESIGN.balance.FAR'] = 0.025
    prob['DESIGN.balance.W'] = 100.
    prob['DESIGN.balance.lpt_PR'] = 4.0
    prob['DESIGN.balance.hpt_PR'] = 3.0
    prob['DESIGN.fc.balance.Pt'] = 5.2
    prob['DESIGN.fc.balance.Tt'] = 440.0


    for pt in ['OD_full_pwr', 'OD_part_pwr']:

        # initial guesses
        prob[pt+'.balance.FAR'] = 0.02467
        prob[pt+'.balance.W'] = 300
        prob[pt+'.balance.BPR'] = 5.105
        prob[pt+'.balance.lp_Nmech'] = 5000
        prob[pt+'.balance.hp_Nmech'] = 15000
        prob[pt+'.hpt.PR'] = 3.
        prob[pt+'.lpt.PR'] = 4.
        prob[pt+'.fan.map.RlineMap'] = 2.0
        prob[pt+'.lpc.map.RlineMap'] = 2.0
        prob[pt+'.hpc.map.RlineMap'] = 2.0

    st = time.time()

    prob.set_solver_print(level=-1)
    prob.set_solver_print(level=2, depth=1)

    flight_env = [(0.8, 35000), (0.7, 35000), (0.55, 35000), (0.46, 35000), (0.4, 35000),
                  (0.4, 20000), (0.6, 20000), (0.8, 20000),
                  (0.8, 10000), (0.6, 10000), (0.4, 10000), (0.2, 10000), (0.001, 10000),
                  (.001, 1000), (0.2, 1000), (0.4, 1000), (0.6, 1000),
                  (0.6, 0), (0.4, 0), (0.2, 0), (0.001, 0)]

    viewer_file = open('hbtf_view.out', 'w')
    first_pass = True
    for MN, alt in flight_env:

        # NOTE: You never change the MN,alt for the
        # design point because that is a fixed reference condition.

        print('***'*10)
        print(f'* MN: {MN}, alt: {alt}')
        print('***'*10)
        prob['OD_full_pwr.fc.MN'] = MN
        prob['OD_full_pwr.fc.alt'] = alt

        prob['OD_part_pwr.fc.MN'] = MN
        prob['OD_part_pwr.fc.alt'] = alt

        for PC in [1, 0.9, 0.8, .7]:
            print(f'## PC = {PC}')
            prob['OD_part_pwr.PC'] = PC
            prob.run_model()

            if first_pass:
                viewer(prob, 'DESIGN', file=viewer_file)
                first_pass = False
            viewer(prob, 'OD_part_pwr', file=viewer_file)

        # run throttle back up to full power
        for PC in [1, 0.85]:
            prob['OD_part_pwr.PC'] = PC
            prob.run_model()



    print()
    print("Run time", time.time() - st)



from __future__ import annotations

import runpy

import pytest

from pycycle_mcp import main as server_main


def test_python_m_entrypoint_exits_with_main_return_code(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(server_main, "main", lambda: 7)

    with pytest.raises(SystemExit) as exc:
        runpy.run_module("pycycle_mcp.__main__", run_name="__main__")

    assert exc.value.code == 7



"""Error helpers for the pyCycle MCP server.

Provides a typed container for MCP errors and helper utilities to keep
responses consistent across tools.
"""

from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class MCPError:
    """Structured MCP error payload.

    Attributes:
        error_type: High-level error category.
        message: Human readable description of the error.
        details: Optional extra context.
    """

    error_type: str
    message: str
    details: object | None = None

    def to_response(self) -> dict[str, object]:
        """Convert the error to a response dictionary."""

        return {
            "error": {
                "type": self.error_type,
                "message": self.message,
                "details": self.details,
            }
        }


def error_response(error_type: str, message: str, details: object | None = None) -> dict[str, object]:
    """Return a standardized MCP error response.

    Args:
        error_type: High-level error category.
        message: Human readable description.
        details: Optional extra context for debugging.

    Returns:
        A mapping that matches the MCP error envelope.
    """

    return MCPError(error_type=error_type, message=message, details=details).to_response()


def to_error(err: Exception) -> dict[str, object]:
    """Convert an exception into an MCP error response."""

    return error_response(error_type=err.__class__.__name__, message=str(err))



"""Derivative computation tools."""

from __future__ import annotations

from typing import cast

from ..errors import error_response, to_error
from ..session_manager import session_manager
from ..utils import error_on_missing_session


def _format_by_pair(
    jacobian: dict[tuple[str, str], object],
) -> dict[str, dict[str, object]]:
    formatted: dict[str, dict[str, object]] = {}
    for (of, wrt), value in jacobian.items():
        formatted.setdefault(of, {})[wrt] = value.tolist() if hasattr(value, "tolist") else value
    return formatted


def compute_totals(payload: dict[str, object]) -> dict[str, object]:
    """Compute total derivatives using OpenMDAO."""

    session_id = payload.get("session_id")
    of: list[str] = payload.get("of", [])  # type: ignore[assignment]
    wrt: list[str] = payload.get("wrt", [])  # type: ignore[assignment]
    return_format = payload.get("return_format", "by_pair")

    if not session_id:
        return error_response("ValidationError", "session_id is required")
    if not of or not wrt:
        return error_response("ValidationError", "of and wrt must be provided")

    try:
        problem, _ = session_manager.get(str(session_id))
        totals = problem.compute_totals(of=of, wrt=wrt)

        if return_format == "dense":
            data = []
            for output_name in of:
                row = []
                for input_name in wrt:
                    cell = totals[(output_name, input_name)]
                    row.append(cell.tolist() if hasattr(cell, "tolist") else cell)
                data.append(row)
            jacobian: dict[str, object] = {"of": of, "wrt": wrt, "data": data}
        else:
            jacobian = cast(dict[str, object], _format_by_pair(totals))

        return {"jacobian": jacobian, "messages": ["Totals computed"]}
    except KeyError as exc:
        return error_on_missing_session(str(session_id), exc)
    except Exception as exc:  # pragma: no cover
        return to_error(exc)



"""Tools for creating and summarizing pyCycle/OpenMDAO models."""

from __future__ import annotations

import logging
from collections.abc import Callable

from ..errors import error_response, to_error
from ..session_manager import session_manager
from ..types import CycleProblem
from ..utils import (
    error_on_missing_session,
    load_callable,
    select_interesting_variables,
)

LOGGER = logging.getLogger(__name__)

INTERESTING_INPUT_KEYWORDS = ["mach", "alt", "pr", "turbine", "throttle"]
INTERESTING_OUTPUT_KEYWORDS = ["fn", "fnet", "thrust", "tsfc", "power", "eff"]


CycleBuilder = Callable[[], object]


def _resolve_builtin_cycle(cycle_type: str) -> tuple[str, CycleBuilder]:
    """Resolve a built-in cycle type to a callable builder.

    NASA pyCycle provides engine components (Compressor, Turbine, etc.)
    rather than ready-made engine classes. We ship bundled cycle definitions
    that wire those components into complete engine models.
    """

    try:
        import pycycle.api as pyc_api  # noqa: F401 – verify pycycle is installed
    except Exception as exc:
        raise ImportError("pycycle (om-pycycle) is required for built-in cycle types") from exc

    from ..cycles.high_bypass_turbofan import HBTF
    from ..cycles.simple_turbojet import Turbojet

    mapping: dict[str, CycleBuilder] = {
        "turbofan": lambda: HBTF(thermo_method="CEA"),
        "turbojet": lambda: Turbojet(),
    }
    builder = mapping.get(cycle_type)
    if builder is None:
        supported = ", ".join(sorted(mapping.keys()))
        raise ValueError(f"Unsupported cycle type: {cycle_type}. Supported: {supported}")
    return cycle_type, builder


def _build_problem(builder: CycleBuilder, mode: str, options: dict[str, object]) -> tuple[CycleProblem, str]:
    try:
        from openmdao.api import Problem
    except Exception as exc:
        raise ImportError("openmdao is required to build cycle models") from exc

    problem: CycleProblem = Problem()
    model = builder()
    problem.model = model  # type: ignore[assignment]

    problem.setup(check=False)

    # Apply default design-point values for built-in cycles
    _apply_design_defaults(problem, model, mode)

    for key, value in options.items():
        try:
            problem.set_val(key, value)
        except Exception as exc:
            LOGGER.debug("Could not set %s=%s: %s", key, value, exc)

    return problem, type(model).__name__


def _apply_design_defaults(problem: CycleProblem, model: object, mode: str) -> None:
    """Set sensible design-point values so the model can run out of the box."""
    from ..cycles.high_bypass_turbofan import HBTF
    from ..cycles.simple_turbojet import Turbojet

    if isinstance(model, HBTF):
        problem.set_val("fan.PR", 1.685)
        problem.set_val("fan.eff", 0.8948)
        problem.set_val("lpc.PR", 1.935)
        problem.set_val("lpc.eff", 0.9243)
        problem.set_val("hpc.PR", 9.369)
        problem.set_val("hpc.eff", 0.8707)
        problem.set_val("hpt.eff", 0.8888)
        problem.set_val("lpt.eff", 0.8996)
        problem.set_val("fc.alt", 35000.0, units="ft")
        problem.set_val("fc.MN", 0.8)
        problem.set_val("T4_MAX", 2857.0, units="degR")
        problem.set_val("Fn_DES", 5900.0, units="lbf")
        # Initial guesses for Newton solver
        problem["balance.FAR"] = 0.025
        problem["balance.W"] = 100.0
        problem["balance.lpt_PR"] = 4.0
        problem["balance.hpt_PR"] = 3.0
        problem["fc.balance.Pt"] = 5.2
        problem["fc.balance.Tt"] = 440.0

    elif isinstance(model, Turbojet):
        problem.set_val("fc.alt", 0.0, units="ft")
        problem.set_val("fc.MN", 0.000001)
        problem.set_val("balance.Fn_target", 11800.0, units="lbf")
        problem.set_val("balance.T4_target", 2370.0, units="degR")
        problem.set_val("comp.PR", 13.5)
        problem.set_val("comp.eff", 0.83)
        problem.set_val("turb.eff", 0.86)
        problem["balance.FAR"] = 0.0175
        problem["balance.W"] = 168.0
        problem["balance.turb_PR"] = 4.46
        problem["fc.balance.Pt"] = 14.696
        problem["fc.balance.Tt"] = 518.67


def _extract_variables(problem: CycleProblem, target: str) -> list[tuple[str, dict[str, object]]]:
    if target == "inputs":
        items = problem.model.list_inputs(prom_name=True, out_stream=None)
    else:
        items = problem.model.list_outputs(prom_name=True, out_stream=None)
    return [(name, meta) for name, meta in items]


def _summarize_variables(
    problem: CycleProblem,
) -> tuple[list[dict[str, object]], list[dict[str, object]]]:
    inputs = _extract_variables(problem, "inputs")
    outputs = _extract_variables(problem, "outputs")

    interesting_inputs = select_interesting_variables(inputs, INTERESTING_INPUT_KEYWORDS)
    interesting_outputs = select_interesting_variables(outputs, INTERESTING_OUTPUT_KEYWORDS)

    def _render(names: list[str], source: list[tuple[str, dict[str, object]]]) -> list[dict[str, object]]:
        rendered: list[dict[str, object]] = []
        metadata_map = {name: meta for name, meta in source}
        for name in names:
            meta = metadata_map.get(name, {})
            rendered.append({"name": name, "units": meta.get("units"), "desc": meta.get("desc")})
        return rendered

    return _render(interesting_inputs, inputs), _render(interesting_outputs, outputs)


def create_cycle_model(payload: dict[str, object]) -> dict[str, object]:
    """Instantiate a pyCycle/OpenMDAO Problem for a specified engine cycle."""

    cycle_type = payload.get("cycle_type")
    mode = payload.get("mode")
    options: dict[str, object] = payload.get("options", {})  # type: ignore[assignment]
    cycle_module_path = payload.get("cycle_module_path")

    if not cycle_type or not mode:
        return error_response("ValidationError", "cycle_type and mode are required")

    try:
        if cycle_type == "custom":
            if not cycle_module_path:
                raise ValueError("cycle_module_path is required for custom cycles")
            builder = load_callable(str(cycle_module_path))
            model_name = getattr(builder, "__name__", cycle_module_path)
        else:
            model_name, builder = _resolve_builtin_cycle(str(cycle_type))

        problem, resolved_name = _build_problem(builder=builder, mode=str(mode), options=options)
        session_id = session_manager.create_session(problem=problem, meta={"mode": mode, "options": options})
        top_inputs, top_outputs = _summarize_variables(problem)

        return {
            "session_id": session_id,
            "model_name": resolved_name or model_name,
            "top_promoted_inputs": top_inputs,
            "top_promoted_outputs": top_outputs,
        }
    except Exception as exc:  # pragma: no cover - handled in tests via fake errors
        LOGGER.error("Failed to create cycle model: %s", exc)
        return to_error(exc)


def close_cycle_model(payload: dict[str, object]) -> dict[str, object]:
    """Close a pyCycle session and free resources."""

    session_id = payload.get("session_id")
    if not session_id:
        return error_response("ValidationError", "session_id is required")

    try:
        session_manager.close(str(session_id))
        return {"success": True}
    except Exception as exc:  # pragma: no cover
        return to_error(exc)


def get_cycle_summary(payload: dict[str, object]) -> dict[str, object]:
    """Return a succinct summary of the current cycle model."""

    session_id = payload.get("session_id")
    if not session_id:
        return error_response("ValidationError", "session_id is required")

    try:
        problem, meta = session_manager.get(str(session_id))
        mode = meta.get("mode")
        options = meta.get("options", {})
        inputs = _extract_variables(problem, "inputs")
        outputs = _extract_variables(problem, "outputs")

        def _populate(
            entries: list[tuple[str, dict[str, object]]],
        ) -> list[dict[str, object]]:
            rendered: list[dict[str, object]] = []
            for name, meta_entry in entries:
                rendered.append(
                    {
                        "name": name,
                        "units": meta_entry.get("units"),
                        "desc": meta_entry.get("desc"),
                        "current_value": meta_entry.get("val") or meta_entry.get("value"),
                    }
                )
            return rendered

        return {
            "model_name": getattr(problem.model, "name", "cycle"),
            "mode": mode,
            "options": options,
            "key_inputs": _populate(inputs),
            "key_outputs": _populate(outputs),
        }
    except KeyError as exc:
        return error_on_missing_session(str(session_id), exc)
    except Exception as exc:  # pragma: no cover
        return to_error(exc)



from __future__ import annotations

import logging

from pycycle_mcp import runtime


class _DummyServer:
    def __init__(self) -> None:
        self.calls: list[dict[str, object]] = []

    def run(self, **kwargs: object) -> None:
        self.calls.append(kwargs)


def test_run_server_passes_settings_to_fastmcp() -> None:
    server = _DummyServer()
    settings = runtime.ServerSettings(
        transport="streamable-http",
        host="0.0.0.0",
        port=9000,
        show_banner=True,
    )

    runtime.run_server(server, settings)

    assert server.calls == [
        {
            "transport": "streamable-http",
            "host": "0.0.0.0",
            "port": 9000,
            "show_banner": True,
        }
    ]


def test_configure_logging_uses_basic_config(monkeypatch) -> None:
    captured: dict[str, object] = {}

    def fake_basic_config(**kwargs: object) -> None:
        captured.update(kwargs)

    monkeypatch.setattr(logging, "basicConfig", fake_basic_config)

    runtime.configure_logging("DEBUG")

    assert captured["level"] == "DEBUG"
    assert "%(asctime)s" in str(captured["format"])



"""Runtime utilities for configuring and running the FastMCP server."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Literal

from fastmcp.server import FastMCP


@dataclass
class ServerSettings:
    """Transport and networking configuration for FastMCP."""

    transport: Literal["stdio", "http", "sse", "streamable-http"] = "stdio"
    host: str = "127.0.0.1"
    port: int = 8000
    show_banner: bool = False


def configure_logging(level: int | str = logging.INFO) -> None:
    """Initialize logging for CLI and server runs."""

    logging.basicConfig(
        level=level,
        format="%(asctime)s %(levelname)s [%(name)s] %(message)s",
    )


def run_server(server: FastMCP, settings: ServerSettings) -> None:
    """Start the FastMCP server with the provided settings."""

    server.run(
        transport=settings.transport,
        host=settings.host,
        port=settings.port,
        show_banner=settings.show_banner,
    )



"""Tool collection for the pyCycle MCP server."""

from .create_model import close_cycle_model, create_cycle_model, get_cycle_summary
from .derivatives import compute_totals
from .execution import run_cycle
from .ping import PingRequest, PingResponse, ping
from .sweep import sweep_inputs
from .variables import get_outputs, list_variables, set_inputs

__all__ = [
    "create_cycle_model",
    "close_cycle_model",
    "get_cycle_summary",
    "list_variables",
    "set_inputs",
    "get_outputs",
    "run_cycle",
    "sweep_inputs",
    "compute_totals",
    "ping",
    "PingRequest",
    "PingResponse",
]



"""Shared typing protocol definitions for cycle problems."""

from __future__ import annotations

from typing import Protocol, runtime_checkable


@runtime_checkable
class CycleModel(Protocol):
    name: str

    def list_inputs(
        self, prom_name: bool = True, out_stream: object | None = None
    ) -> list[tuple[str, dict[str, object]]]:  # pragma: no cover - protocol
        ...

    def list_outputs(
        self, prom_name: bool = True, out_stream: object | None = None
    ) -> list[tuple[str, dict[str, object]]]:  # pragma: no cover - protocol
        ...


@runtime_checkable
class CycleProblem(Protocol):
    model: CycleModel
    iter_count: int | None
    values: dict[str, object] | None

    def set_val(
        self, name: str | None = None, value: object | None = None, **kwargs: object
    ) -> None:  # pragma: no cover
        ...

    def get_val(self, name: str, units: str | None = None) -> object:  # pragma: no cover
        ...

    def run_model(self) -> None:  # pragma: no cover
        ...

    def run_driver(self) -> None:  # pragma: no cover
        ...

    def setup(self, check: bool = False) -> None:  # pragma: no cover
        ...

    def set_solver_print(self, level: int = 0) -> None:  # pragma: no cover
        ...

    def __setitem__(self, name: str, value: object) -> None:  # pragma: no cover
        ...

    def compute_totals(self, of: list[str], wrt: list[str]) -> dict[tuple[str, str], object]:  # pragma: no cover
        ...



from __future__ import annotations

import sys
import types

import pytest

from pycycle_mcp.tools import create_model


class _RecorderProblem:
    def __init__(self) -> None:
        self.model: object | None = None
        self.setup_calls: list[bool] = []
        self.set_calls: list[tuple[str, object, object | None]] = []
        self.item_calls: dict[str, object] = {}

    def setup(self, check: bool = False) -> None:
        self.setup_calls.append(check)

    def set_val(
        self,
        name: str | None = None,
        value: object | None = None,
        **kwargs: object,
    ) -> None:
        if name == "bad":
            raise RuntimeError("bad set")
        units = kwargs.get("units")
        if name is not None:
            self.set_calls.append((name, value, units))

    def __setitem__(self, name: str, value: object) -> None:
        self.item_calls[name] = value


def _install_fake_cycle_modules(monkeypatch: pytest.MonkeyPatch) -> tuple[type[object], type[object]]:
    pycycle_module = types.ModuleType("pycycle")
    pycycle_api_module = types.ModuleType("pycycle.api")
    pycycle_module.api = pycycle_api_module

    hb_module = types.ModuleType("pycycle_mcp.cycles.high_bypass_turbofan")

    class HBTF:
        def __init__(self, thermo_method: str | None = None) -> None:
            self.thermo_method = thermo_method

    hb_module.HBTF = HBTF

    tj_module = types.ModuleType("pycycle_mcp.cycles.simple_turbojet")

    class Turbojet:
        pass

    tj_module.Turbojet = Turbojet

    monkeypatch.setitem(sys.modules, "pycycle", pycycle_module)
    monkeypatch.setitem(sys.modules, "pycycle.api", pycycle_api_module)
    monkeypatch.setitem(sys.modules, "pycycle_mcp.cycles.high_bypass_turbofan", hb_module)
    monkeypatch.setitem(sys.modules, "pycycle_mcp.cycles.simple_turbojet", tj_module)

    return HBTF, Turbojet


def test_resolve_builtin_cycle_import_error(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delitem(sys.modules, "pycycle", raising=False)
    monkeypatch.delitem(sys.modules, "pycycle.api", raising=False)

    with pytest.raises(ImportError):
        create_model._resolve_builtin_cycle("turbofan")


def test_resolve_builtin_cycle_success_and_unsupported(monkeypatch: pytest.MonkeyPatch) -> None:
    hbtf_cls, turbojet_cls = _install_fake_cycle_modules(monkeypatch)

    cycle_type, turbofan_builder = create_model._resolve_builtin_cycle("turbofan")
    assert cycle_type == "turbofan"
    turbofan = turbofan_builder()
    assert isinstance(turbofan, hbtf_cls)
    assert turbofan.thermo_method == "CEA"

    cycle_type, turbojet_builder = create_model._resolve_builtin_cycle("turbojet")
    assert cycle_type == "turbojet"
    assert isinstance(turbojet_builder(), turbojet_cls)

    with pytest.raises(ValueError):
        create_model._resolve_builtin_cycle("unsupported")


def test_apply_design_defaults_for_both_builtin_cycle_types(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    hbtf_cls, turbojet_cls = _install_fake_cycle_modules(monkeypatch)

    hbtf_problem = _RecorderProblem()
    create_model._apply_design_defaults(hbtf_problem, hbtf_cls(), mode="design")
    assert any(name == "fan.PR" for name, _, _ in hbtf_problem.set_calls)
    assert hbtf_problem.item_calls["balance.FAR"] == 0.025

    turbojet_problem = _RecorderProblem()
    create_model._apply_design_defaults(turbojet_problem, turbojet_cls(), mode="design")
    assert any(name == "comp.PR" for name, _, _ in turbojet_problem.set_calls)
    assert turbojet_problem.item_calls["balance.turb_PR"] == 4.46


class _FakeModel:
    pass


class _OpenMDAOProblem(_RecorderProblem):
    pass


def test_build_problem_import_error_when_openmdao_missing(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delitem(sys.modules, "openmdao", raising=False)
    monkeypatch.delitem(sys.modules, "openmdao.api", raising=False)

    with pytest.raises(ImportError):
        create_model._build_problem(lambda: _FakeModel(), "design", {})


def test_build_problem_success_and_option_failures(monkeypatch: pytest.MonkeyPatch) -> None:
    openmdao_module = types.ModuleType("openmdao")
    openmdao_api_module = types.ModuleType("openmdao.api")
    openmdao_api_module.Problem = _OpenMDAOProblem
    openmdao_module.api = openmdao_api_module

    monkeypatch.setitem(sys.modules, "openmdao", openmdao_module)
    monkeypatch.setitem(sys.modules, "openmdao.api", openmdao_api_module)
    monkeypatch.setattr(create_model, "_apply_design_defaults", lambda problem, model, mode: None)

    problem, model_name = create_model._build_problem(
        lambda: _FakeModel(),
        mode="design",
        options={"good": 1.0, "bad": 2.0},
    )

    assert model_name == "_FakeModel"
    assert problem.setup_calls == [False]
    assert ("good", 1.0, None) in problem.set_calls


def test_close_cycle_model_success_path() -> None:
    response = create_model.close_cycle_model({"session_id": "anything"})
    assert response == {"success": True}



"""Typed request and response payloads for FastMCP tool bindings.

The models defined here mirror the existing dictionary-based tool contracts
but provide explicit JSON Schema generation for FastMCP consumers.
"""

from __future__ import annotations

from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, Field


class ToolError(BaseModel):
    """Structured error envelope shared by all tool responses."""

    type: str
    message: str
    details: Any | None = None

    model_config = ConfigDict(extra="ignore")


class BaseResponse(BaseModel):
    """Base response wrapper that carries optional error information."""

    error: ToolError | None = None

    model_config = ConfigDict(extra="ignore")


class NamedVariable(BaseModel):
    """Common variable metadata representation."""

    name: str
    units: str | None = None
    desc: str | None = None
    current_value: Any | None = None
    io: str | None = None
    promoted: bool | None = None
    shape: Any | None = None
    value: Any | None = None

    model_config = ConfigDict(extra="ignore")


class CreateCycleModelRequest(BaseModel):
    """Payload for creating a new pyCycle/OpenMDAO model."""

    cycle_type: str
    mode: str
    options: dict[str, Any] = Field(default_factory=dict)
    cycle_module_path: str | None = None

    model_config = ConfigDict(extra="forbid")


class CreateCycleModelResponse(BaseResponse):
    """Response returned after creating a cycle model."""

    session_id: str | None = None
    model_name: str | None = None
    top_promoted_inputs: list[NamedVariable] = Field(default_factory=list)
    top_promoted_outputs: list[NamedVariable] = Field(default_factory=list)


class CloseCycleModelResponse(BaseResponse):
    """Response from closing a cycle model session."""

    success: bool | None = None


class CycleSummaryResponse(BaseResponse):
    """Summary of a model's configuration and key variables."""

    model_name: str | None = None
    mode: str | None = None
    options: dict[str, Any] = Field(default_factory=dict)
    key_inputs: list[NamedVariable] = Field(default_factory=list)
    key_outputs: list[NamedVariable] = Field(default_factory=list)


class ListVariablesRequest(BaseModel):
    """Payload for listing variables."""

    session_id: str
    kind: Literal["inputs", "outputs", "both"] = "both"
    promoted_only: bool = True
    name_filter: str | None = None
    max_variables: int = 200

    model_config = ConfigDict(extra="forbid")


class ListVariablesResponse(BaseResponse):
    """Response containing variable metadata entries."""

    variables: list[NamedVariable] = Field(default_factory=list)


class SetInputsRequest(BaseModel):
    """Payload for setting one or more input variables."""

    session_id: str
    values: dict[str, Any]
    allow_missing: bool = False

    model_config = ConfigDict(extra="forbid")


class SetInputsResponse(BaseResponse):
    """Response from setting inputs."""

    updated: list[str] = Field(default_factory=list)
    skipped: list[dict[str, Any]] = Field(default_factory=list)


class GetOutputsRequest(BaseModel):
    """Payload for retrieving one or more outputs."""

    session_id: str
    names: list[str]
    allow_missing: bool = False

    model_config = ConfigDict(extra="forbid")


class GetOutputsResponse(BaseResponse):
    """Response carrying requested outputs."""

    values: dict[str, Any] = Field(default_factory=dict)
    missing: list[str] = Field(default_factory=list)


class RunCycleRequest(BaseModel):
    """Payload for executing a cycle run."""

    session_id: str
    outputs_of_interest: list[str] = Field(default_factory=list)
    use_driver: bool = False

    model_config = ConfigDict(extra="forbid")


class RunCycleResponse(BaseResponse):
    """Response from running a cycle model."""

    success: bool | None = None
    converged: Any | None = None
    iterations: int | None = None
    outputs: dict[str, Any] = Field(default_factory=dict)
    residual_norm: Any | None = None
    messages: list[str] = Field(default_factory=list)


class SweepVariable(BaseModel):
    """Single sweep variable specification."""

    name: str
    values: list[Any]

    model_config = ConfigDict(extra="forbid")


class SweepInputsRequest(BaseModel):
    """Payload describing a sweep run."""

    session_id: str
    sweep: list[SweepVariable]
    outputs_of_interest: list[str] = Field(default_factory=list)
    use_driver: bool = False
    skip_on_failure: bool = True

    model_config = ConfigDict(extra="forbid")


class SweepResult(BaseModel):
    """Result for a single sweep iteration."""

    inputs: dict[str, Any] = Field(default_factory=dict)
    success: bool
    outputs: dict[str, Any] = Field(default_factory=dict)
    error_message: str | None = None

    model_config = ConfigDict(extra="ignore")


class SweepInputsResponse(BaseResponse):
    """Response for sweep execution."""

    results: list[SweepResult] = Field(default_factory=list)


class ComputeTotalsRequest(BaseModel):
    """Payload for computing total derivatives."""

    session_id: str
    of: list[str]
    wrt: list[str]
    return_format: Literal["by_pair", "dense"] = "by_pair"

    model_config = ConfigDict(extra="forbid")


class ComputeTotalsResponse(BaseResponse):
    """Response carrying jacobian data."""

    jacobian: dict[str, Any] = Field(default_factory=dict)
    messages: list[str] = Field(default_factory=list)



"""Utility helpers for the pyCycle MCP server."""

from __future__ import annotations

import importlib
import logging
from collections.abc import Callable, Iterable, Sequence
from typing import cast

from .errors import error_response

LOGGER = logging.getLogger(__name__)


def load_callable(path: str) -> Callable[[], object]:
    """Load a callable from a fully-qualified import path."""

    module_path, _, attr = path.rpartition(".")
    if not module_path:
        raise ImportError(f"Invalid import path: {path}")
    module = importlib.import_module(module_path)
    target = getattr(module, attr)
    if not callable(target):
        raise TypeError(f"Target at {path} is not callable")
    return cast(Callable[[], object], target)


def error_on_missing_session(session_id: str, exc: Exception) -> dict[str, object]:
    """Standardize unknown session errors."""

    return error_response(
        error_type=exc.__class__.__name__,
        message=f"Session '{session_id}' not found",
        details=str(exc),
    )


def select_interesting_variables(variables: list[tuple[str, dict[str, object]]], keywords: Iterable[str]) -> list[str]:
    """Select promoted names containing any keyword in a case-insensitive fashion."""

    selected: list[str] = []
    for name, _ in variables:
        lowered = name.lower()
        if any(keyword.lower() in lowered for keyword in keywords):
            selected.append(name)
    return selected


def render_variable_entry(name: str, metadata: dict[str, object], io: str) -> dict[str, object]:
    """Format an OpenMDAO variable metadata entry."""

    value = metadata.get("value") if "value" in metadata else metadata.get("val")
    return {
        "name": name,
        "io": io,
        "promoted": metadata.get("promoted_name", name) == name,
        "units": metadata.get("units"),
        "shape": _normalize_shape(metadata.get("shape")),
        "desc": metadata.get("desc"),
        "value": value,
    }


def _normalize_shape(shape: object | None) -> object | None:
    if shape is None:
        return None
    if isinstance(shape, Sequence) and not isinstance(shape, (str, bytes)):
        return list(shape)
    return shape


def ordered_cartesian_product(values: Sequence[Sequence[object]]) -> list[list[object]]:
    """Return nested-loop ordering for a Cartesian product."""

    if not values:
        return []
    results: list[list[object]] = [[]]
    for group in values:
        next_results: list[list[object]] = []
        for prefix in results:
            for item in group:
                next_results.append(prefix + [item])
        results = next_results
    return results



from __future__ import annotations

from typing import cast

from pycycle_mcp.session_manager import SessionManager
from pycycle_mcp.types import CycleProblem

from .conftest import DummyProblem


def test_session_lifecycle() -> None:
    manager = SessionManager()
    problem_obj = cast(CycleProblem, DummyProblem())
    session_id = manager.create_session(problem=problem_obj, meta={"mode": "design"})
    problem, meta = manager.get(session_id)

    assert problem is problem_obj
    assert meta["mode"] == "design"

    manager.close(session_id)
    try:
        manager.get(session_id)
    except KeyError:
        assert True
    else:  # pragma: no cover - safety
        raise AssertionError("Expected session to be removed")



"""Factory for the FastMCP server exposing pyCycle tools."""

from __future__ import annotations

import logging
from collections.abc import Callable
from typing import Any, Literal, TypedDict, TypeVar, Unpack, cast

from fastmcp.server import FastMCP
from fastmcp.tools.tool import FunctionTool
from mcp.types import ToolAnnotations
from pydantic import BaseModel

from . import tools
from .schemas import (
    CloseCycleModelResponse,
    ComputeTotalsRequest,
    ComputeTotalsResponse,
    CreateCycleModelRequest,
    CreateCycleModelResponse,
    CycleSummaryResponse,
    GetOutputsRequest,
    GetOutputsResponse,
    ListVariablesRequest,
    ListVariablesResponse,
    RunCycleRequest,
    RunCycleResponse,
    SetInputsRequest,
    SetInputsResponse,
    SweepInputsRequest,
    SweepInputsResponse,
    SweepVariable,
)
from .tools import PingResponse

__all__ = ["build_server", "tools"]

LOGGER = logging.getLogger(__name__)

ResponseModel = TypeVar("ResponseModel", bound=BaseModel)
ToolFunc = Callable[..., dict[str, Any]]


class ToolKwargs(TypedDict, total=False):
    """Subset of FastMCP tool configuration parameters used by this module."""

    name: str
    title: str | None
    description: str | None
    tags: set[str]
    output_schema: dict[str, Any] | None
    annotations: ToolAnnotations | dict[str, Any] | None


def _tool(server: FastMCP, **kwargs: Unpack[ToolKwargs]) -> Callable[[ToolFunc], ToolFunc]:
    """Typed wrapper around :meth:`FastMCP.tool` to satisfy type checkers."""

    return cast(Callable[[ToolFunc], ToolFunc], server.tool(**kwargs))


def _validated_response(response: dict[str, Any], response_model: type[ResponseModel]) -> dict[str, Any]:
    """Validate and normalize a tool response using a Pydantic model."""

    validated = response_model.model_validate(response)
    return validated.model_dump(exclude_none=True)


def _register_tools(server: FastMCP) -> None:
    """Attach the pyCycle tool implementations to a FastMCP instance."""

    server.add_tool(
        FunctionTool.from_function(
            tools.ping,
            name="ping",
            description="Simple healthcheck for the pyCycle MCP server.",
            annotations=ToolAnnotations(title="Ping", readOnlyHint=True),
            output_schema=PingResponse.model_json_schema(),
        )
    )

    @_tool(
        server,
        name="create_cycle_model",
        description="Instantiate a pyCycle/OpenMDAO Problem for a specified engine cycle.",
        tags={"pycycle", "model"},
        output_schema=CreateCycleModelResponse.model_json_schema(),
        annotations=ToolAnnotations(title="Create cycle model"),
    )
    def create_cycle_model_tool(
        cycle_type: str,
        mode: str,
        options: dict[str, Any] | None = None,
        cycle_module_path: str | None = None,
    ) -> dict[str, Any]:
        request = CreateCycleModelRequest(
            cycle_type=cycle_type,
            mode=mode,
            options=options or {},
            cycle_module_path=cycle_module_path,
        )
        response = tools.create_model.create_cycle_model(request.model_dump(exclude_none=True))
        return _validated_response(response, CreateCycleModelResponse)

    @_tool(
        server,
        name="close_cycle_model",
        description="Close a pyCycle session and free resources.",
        tags={"pycycle", "session"},
        output_schema=CloseCycleModelResponse.model_json_schema(),
        annotations=ToolAnnotations(title="Close cycle model", destructiveHint=True),
    )
    def close_cycle_model_tool(session_id: str) -> dict[str, Any]:
        response = tools.create_model.close_cycle_model({"session_id": session_id})
        return _validated_response(response, CloseCycleModelResponse)

    @_tool(
        server,
        name="get_cycle_summary",
        description="Return a succinct summary of the current cycle model.",
        tags={"pycycle", "summary"},
        output_schema=CycleSummaryResponse.model_json_schema(),
        annotations=ToolAnnotations(title="Get cycle summary", readOnlyHint=True),
    )
    def get_cycle_summary_tool(session_id: str) -> dict[str, Any]:
        response = tools.create_model.get_cycle_summary({"session_id": session_id})
        return _validated_response(response, CycleSummaryResponse)

    @_tool(
        server,
        name="list_variables",
        description="List variables in the cycle model.",
        tags={"pycycle", "variables"},
        output_schema=ListVariablesResponse.model_json_schema(),
        annotations=ToolAnnotations(title="List variables", readOnlyHint=True),
    )
    def list_variables_tool(
        session_id: str,
        kind: Literal["inputs", "outputs", "both"] = "both",
        promoted_only: bool = True,
        name_filter: str | None = None,
        max_variables: int = 200,
    ) -> dict[str, Any]:
        request = ListVariablesRequest(
            session_id=session_id,
            kind=kind,
            promoted_only=promoted_only,
            name_filter=name_filter,
            max_variables=max_variables,
        )
        response = tools.variables.list_variables(request.model_dump(exclude_none=True))
        return _validated_response(response, ListVariablesResponse)

    @_tool(
        server,
        name="set_inputs",
        description="Set one or more input variables in the cycle model.",
        tags={"pycycle", "variables"},
        output_schema=SetInputsResponse.model_json_schema(),
        annotations=ToolAnnotations(title="Set inputs"),
    )
    def set_inputs_tool(
        session_id: str,
        values: dict[str, Any],
        allow_missing: bool = False,
    ) -> dict[str, Any]:
        request = SetInputsRequest(session_id=session_id, values=values, allow_missing=allow_missing)
        response = tools.variables.set_inputs(request.model_dump())
        return _validated_response(response, SetInputsResponse)

    @_tool(
        server,
        name="get_outputs",
        description="Fetch values for one or more outputs after a run.",
        tags={"pycycle", "variables"},
        output_schema=GetOutputsResponse.model_json_schema(),
        annotations=ToolAnnotations(title="Get outputs", readOnlyHint=True),
    )
    def get_outputs_tool(
        session_id: str,
        names: list[str],
        allow_missing: bool = False,
    ) -> dict[str, Any]:
        request = GetOutputsRequest(session_id=session_id, names=names, allow_missing=allow_missing)
        response = tools.variables.get_outputs(request.model_dump())
        return _validated_response(response, GetOutputsResponse)

    @_tool(
        server,
        name="run_cycle",
        description="Run the cycle model and return selected outputs.",
        tags={"pycycle", "execution"},
        output_schema=RunCycleResponse.model_json_schema(),
        annotations=ToolAnnotations(title="Run cycle"),
    )
    def run_cycle_tool(
        session_id: str,
        outputs_of_interest: list[str] | None = None,
        use_driver: bool = False,
    ) -> dict[str, Any]:
        request = RunCycleRequest(
            session_id=session_id,
            outputs_of_interest=outputs_of_interest or [],
            use_driver=use_driver,
        )
        response = tools.execution.run_cycle(request.model_dump())
        return _validated_response(response, RunCycleResponse)

    @_tool(
        server,
        name="sweep_inputs",
        description="Perform a parametric sweep over input variables.",
        tags={"pycycle", "sweep"},
        output_schema=SweepInputsResponse.model_json_schema(),
        annotations=ToolAnnotations(title="Sweep inputs"),
    )
    def sweep_inputs_tool(
        session_id: str,
        sweep: list[SweepVariable],
        outputs_of_interest: list[str] | None = None,
        use_driver: bool = False,
        skip_on_failure: bool = True,
    ) -> dict[str, Any]:
        request = SweepInputsRequest(
            session_id=session_id,
            sweep=sweep,
            outputs_of_interest=outputs_of_interest or [],
            use_driver=use_driver,
            skip_on_failure=skip_on_failure,
        )
        response = tools.sweep.sweep_inputs(request.model_dump())
        return _validated_response(response, SweepInputsResponse)

    @_tool(
        server,
        name="compute_totals",
        description="Compute total derivatives using OpenMDAO.",
        tags={"pycycle", "derivatives"},
        output_schema=ComputeTotalsResponse.model_json_schema(),
        annotations=ToolAnnotations(title="Compute totals", readOnlyHint=True),
    )
    def compute_totals_tool(
        session_id: str,
        of: list[str],
        wrt: list[str],
        return_format: Literal["by_pair", "dense"] = "by_pair",
    ) -> dict[str, Any]:
        request = ComputeTotalsRequest(
            session_id=session_id,
            of=of,
            wrt=wrt,
            return_format=return_format,
        )
        response = tools.derivatives.compute_totals(request.model_dump())
        return _validated_response(response, ComputeTotalsResponse)


def build_server() -> FastMCP:
    """Construct a FastMCP server with all pyCycle tools registered."""

    server = FastMCP(
        name="pycycle-mcp",
        instructions=("Expose pyCycle/OpenMDAO utilities for creating and running engine cycle models."),
        strict_input_validation=True,
    )
    _register_tools(server)
    LOGGER.debug("FastMCP server configured")
    return server



from __future__ import annotations

import sys
import types

import pytest

from pycycle_mcp import utils


def test_load_callable_invalid_path_and_non_callable(monkeypatch: pytest.MonkeyPatch) -> None:
    with pytest.raises(ImportError):
        utils.load_callable("not_a_valid_path")

    fake_module = types.ModuleType("tests.fake_module_for_utils")
    fake_module.not_callable = 123
    monkeypatch.setitem(sys.modules, "tests.fake_module_for_utils", fake_module)

    with pytest.raises(TypeError):
        utils.load_callable("tests.fake_module_for_utils.not_callable")


def test_normalize_shape_and_cartesian_empty() -> None:
    assert utils._normalize_shape((2, 3)) == [2, 3]
    assert utils._normalize_shape(5) == 5
    assert utils.ordered_cartesian_product([]) == []



"""Variable management tools for pyCycle models."""

from __future__ import annotations

from ..errors import error_response, to_error
from ..session_manager import session_manager
from ..utils import error_on_missing_session, render_variable_entry


def list_variables(payload: dict[str, object]) -> dict[str, object]:
    """List variables in the cycle model."""

    session_id = payload.get("session_id")
    if not session_id:
        return error_response("ValidationError", "session_id is required")

    kind = payload.get("kind", "both")
    promoted_only = payload.get("promoted_only", True)
    name_filter = payload.get("name_filter")
    max_variables_raw = payload.get("max_variables", 200)
    max_variables = int(max_variables_raw) if isinstance(max_variables_raw, (int, str)) else 200

    try:
        problem, _ = session_manager.get(str(session_id))
        results: list[dict[str, object]] = []

        if kind in ("inputs", "both"):
            for name, metadata in problem.model.list_inputs(prom_name=True, out_stream=None):
                if promoted_only and metadata.get("promoted_name") not in (None, name):
                    continue
                if name_filter and str(name_filter).lower() not in name.lower():
                    continue
                results.append(render_variable_entry(name, metadata, "input"))

        if kind in ("outputs", "both"):
            for name, metadata in problem.model.list_outputs(prom_name=True, out_stream=None):
                if promoted_only and metadata.get("promoted_name") not in (None, name):
                    continue
                if name_filter and str(name_filter).lower() not in name.lower():
                    continue
                results.append(render_variable_entry(name, metadata, "output"))

        return {"variables": results[:max_variables]}
    except KeyError as exc:
        return error_on_missing_session(str(session_id), exc)
    except Exception as exc:  # pragma: no cover
        return to_error(exc)


def set_inputs(payload: dict[str, object]) -> dict[str, object]:
    """Set one or more input variables in the cycle model."""

    session_id = payload.get("session_id")
    values: dict[str, object] = payload.get("values", {})  # type: ignore[assignment]
    allow_missing = bool(payload.get("allow_missing", False))

    if not session_id:
        return error_response("ValidationError", "session_id is required")
    if not values:
        return error_response("ValidationError", "values must contain at least one entry")

    try:
        problem, _ = session_manager.get(str(session_id))
        updated: list[str] = []
        skipped: list[dict[str, str]] = []

        for name, value in values.items():
            try:
                problem.set_val(name, value)
                updated.append(name)
            except Exception as exc:
                if allow_missing:
                    skipped.append({"name": name, "reason": str(exc)})
                else:
                    return to_error(exc)

        return {"updated": updated, "skipped": skipped}
    except KeyError as exc:
        return error_on_missing_session(str(session_id), exc)
    except Exception as exc:  # pragma: no cover
        return to_error(exc)


def get_outputs(payload: dict[str, object]) -> dict[str, object]:
    """Fetch values for one or more outputs after a run."""

    session_id = payload.get("session_id")
    names: list[str] = payload.get("names", [])  # type: ignore[assignment]
    allow_missing = bool(payload.get("allow_missing", False))

    if not session_id:
        return error_response("ValidationError", "session_id is required")
    if not names:
        return error_response("ValidationError", "names must contain at least one entry")

    try:
        problem, _ = session_manager.get(str(session_id))
        values: dict[str, object] = {}
        missing: list[str] = []

        for name in names:
            try:
                current = problem.get_val(name)
                values[name] = current.item() if hasattr(current, "item") else current
            except Exception:
                if allow_missing:
                    missing.append(name)
                else:
                    return error_response("LookupError", f"Unknown output: {name}")

        return {"values": values, "missing": missing}
    except KeyError as exc:
        return error_on_missing_session(str(session_id), exc)
    except Exception as exc:  # pragma: no cover
        return to_error(exc)



from __future__ import annotations

from collections.abc import Callable
from typing import cast

import pytest

from pycycle_mcp.session_manager import session_manager
from pycycle_mcp.tools import create_model
from pycycle_mcp.types import CycleProblem

from .conftest import DummyModel, DummyProblem


def dummy_builder() -> DummyModel:
    return DummyModel(
        inputs=[("Mach", {"units": "", "desc": "Mach"})],
        outputs=[("Fn", {"units": "lbf"})],
    )


def test_create_cycle_model_custom(monkeypatch: pytest.MonkeyPatch) -> None:
    def fake_build_problem(
        builder: Callable[[], DummyModel], mode: str, options: dict[str, object]
    ) -> tuple[DummyProblem, str]:
        problem = DummyProblem()
        problem.model = builder()
        return problem, "dummy"

    monkeypatch.setattr(create_model, "_build_problem", fake_build_problem)

    response = create_model.create_cycle_model(
        {
            "cycle_type": "custom",
            "cycle_module_path": "tests.test_create_model.dummy_builder",
            "mode": "design",
            "options": {"Mach": 0.8},
        }
    )

    assert "session_id" in response
    assert response["model_name"] == "dummy"
    assert response["top_promoted_inputs"]
    assert response["top_promoted_outputs"]


def test_get_cycle_summary(monkeypatch: pytest.MonkeyPatch) -> None:
    problem = DummyProblem()
    problem.model.inputs = [("alt", {"units": "ft", "desc": "Altitude", "val": 35000})]
    problem.model.outputs = [("Fn", {"units": "lbf", "desc": "Net thrust", "val": 1000})]
    session_id = session_manager.create_session(
        problem=cast(CycleProblem, problem), meta={"mode": "design", "options": {}}
    )

    response = create_model.get_cycle_summary({"session_id": session_id})
    key_inputs = cast(list[dict[str, object]], response["key_inputs"])
    key_outputs = cast(list[dict[str, object]], response["key_outputs"])
    assert response["model_name"] == "dummy"
    assert response["mode"] == "design"
    assert key_inputs[0]["name"] == "alt"
    assert key_outputs[0]["name"] == "Fn"


def test_create_cycle_model_missing_fields() -> None:
    response = create_model.create_cycle_model({"cycle_type": "turbofan"})
    assert "error" in response



from __future__ import annotations

from typing import cast

from pycycle_mcp.session_manager import session_manager
from pycycle_mcp.tools import derivatives, sweep
from pycycle_mcp.types import CycleProblem

from .conftest import DummyProblem


def test_sweep_inputs_success() -> None:
    problem = DummyProblem()
    problem.model.inputs = [("Mach", {"promoted_name": "Mach"})]
    session_id = session_manager.create_session(
        problem=cast(CycleProblem, problem), meta={"mode": "design", "options": {}}
    )

    result = sweep.sweep_inputs(
        {
            "session_id": session_id,
            "sweep": [{"name": "Mach", "values": [0.7, 0.8]}],
            "outputs_of_interest": ["Fn"],
        }
    )

    results = cast(list[dict[str, object]], result["results"])
    assert len(results) == 2
    assert all(bool(entry["success"]) for entry in results)


def test_compute_totals_formats_by_pair() -> None:
    problem = DummyProblem()
    session_id = session_manager.create_session(
        problem=cast(CycleProblem, problem), meta={"mode": "design", "options": {}}
    )

    result = derivatives.compute_totals(
        {
            "session_id": session_id,
            "of": ["Fn"],
            "wrt": ["Mach"],
            "return_format": "by_pair",
        }
    )

    jacobian = cast(dict[str, dict[str, object]], result["jacobian"])
    assert "Fn" in jacobian
    assert jacobian["Fn"]["Mach"] == [[1.0]]



"""Lightweight healthcheck tool for the pyCycle MCP server."""

from __future__ import annotations

from pydantic import BaseModel, Field


class PingRequest(BaseModel):
    """Request payload for the :func:`ping` tool."""

    message: str | None = Field(default=None, description="Optional echo message included in the response.")


class PingResponse(BaseModel):
    """Response payload for the :func:`ping` tool."""

    server: str = Field(default="pycycle-mcp", description="Server identifier.")
    status: str = Field(default="ok", description="Simple health indicator.")
    message: str | None = Field(
        default=None,
        description="Optional echo message provided by the caller, if any.",
    )


def ping(args: PingRequest | None = None) -> PingResponse:
    """Return a simple health response without importing heavy dependencies."""

    args = args or PingRequest()
    return PingResponse(message=args.message)



from __future__ import annotations

import anyio
from pytest import MonkeyPatch

from pycycle_mcp import fastmcp_server


def test_build_server_exposes_schemas() -> None:
    server = fastmcp_server.build_server()

    tool = anyio.run(server.get_tool, "compute_totals")

    assert tool.output_schema is not None
    assert tool.output_schema["type"] == "object"
    assert "jacobian" in tool.output_schema.get("properties", {})


def test_ping_tool_is_registered() -> None:
    server = fastmcp_server.build_server()

    tool = anyio.run(server.get_tool, "ping")
    result = anyio.run(tool.run, {})

    assert result.structured_content is not None
    assert result.structured_content["server"] == "pycycle-mcp"


def test_create_cycle_model_wrapper_returns_structured(
    monkeypatch: MonkeyPatch,
) -> None:
    def fake_create(payload: dict[str, object]) -> dict[str, object]:
        assert payload["cycle_type"] == "custom"
        return {
            "session_id": "abc123",
            "model_name": "demo",
            "top_promoted_inputs": [],
            "top_promoted_outputs": [],
        }

    monkeypatch.setattr(fastmcp_server.tools.create_model, "create_cycle_model", fake_create)
    server = fastmcp_server.build_server()
    tool = anyio.run(server.get_tool, "create_cycle_model")

    result = anyio.run(
        tool.run,
        {
            "cycle_type": "custom",
            "mode": "design",
            "cycle_module_path": "tests.test_create_model.dummy_builder",
            "options": {},
        },
    )

    assert result.structured_content is not None
    assert result.structured_content["session_id"] == "abc123"
    assert result.structured_content["model_name"] == "demo"


def test_all_tool_wrappers_return_validated_structured_content(
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.setattr(
        fastmcp_server.tools.create_model,
        "close_cycle_model",
        lambda payload: {"success": True},
    )
    monkeypatch.setattr(
        fastmcp_server.tools.create_model,
        "get_cycle_summary",
        lambda payload: {
            "model_name": "demo",
            "mode": "design",
            "options": {},
            "key_inputs": [],
            "key_outputs": [],
        },
    )
    monkeypatch.setattr(
        fastmcp_server.tools.variables,
        "list_variables",
        lambda payload: {"variables": []},
    )
    monkeypatch.setattr(
        fastmcp_server.tools.variables,
        "set_inputs",
        lambda payload: {"updated": [], "skipped": []},
    )
    monkeypatch.setattr(
        fastmcp_server.tools.variables,
        "get_outputs",
        lambda payload: {"values": {}, "missing": []},
    )
    monkeypatch.setattr(
        fastmcp_server.tools.execution,
        "run_cycle",
        lambda payload: {"success": True, "outputs": {}, "messages": []},
    )
    monkeypatch.setattr(
        fastmcp_server.tools.sweep,
        "sweep_inputs",
        lambda payload: {"results": []},
    )
    monkeypatch.setattr(
        fastmcp_server.tools.derivatives,
        "compute_totals",
        lambda payload: {"jacobian": {}, "messages": []},
    )

    server = fastmcp_server.build_server()

    close_tool = anyio.run(server.get_tool, "close_cycle_model")
    summary_tool = anyio.run(server.get_tool, "get_cycle_summary")
    list_tool = anyio.run(server.get_tool, "list_variables")
    set_tool = anyio.run(server.get_tool, "set_inputs")
    get_tool = anyio.run(server.get_tool, "get_outputs")
    run_tool = anyio.run(server.get_tool, "run_cycle")
    sweep_tool = anyio.run(server.get_tool, "sweep_inputs")
    totals_tool = anyio.run(server.get_tool, "compute_totals")

    assert anyio.run(close_tool.run, {"session_id": "s"}).structured_content == {"success": True}
    assert anyio.run(summary_tool.run, {"session_id": "s"}).structured_content == {
        "model_name": "demo",
        "mode": "design",
        "options": {},
        "key_inputs": [],
        "key_outputs": [],
    }
    assert anyio.run(
        list_tool.run,
        {"session_id": "s", "kind": "both", "promoted_only": True, "max_variables": 5},
    ).structured_content == {"variables": []}
    assert anyio.run(
        set_tool.run,
        {"session_id": "s", "values": {"Mach": 0.8}, "allow_missing": True},
    ).structured_content == {"updated": [], "skipped": []}
    assert anyio.run(get_tool.run, {"session_id": "s", "names": ["Fn"], "allow_missing": True}).structured_content == {
        "values": {},
        "missing": [],
    }
    assert anyio.run(
        run_tool.run,
        {"session_id": "s", "outputs_of_interest": ["Fn"], "use_driver": False},
    ).structured_content == {"success": True, "outputs": {}, "messages": []}
    assert anyio.run(
        sweep_tool.run,
        {
            "session_id": "s",
            "sweep": [{"name": "Mach", "values": [0.7, 0.8]}],
            "outputs_of_interest": ["Fn"],
            "use_driver": False,
            "skip_on_failure": True,
        },
    ).structured_content == {"results": []}
    assert anyio.run(
        totals_tool.run,
        {"session_id": "s", "of": ["Fn"], "wrt": ["Mach"], "return_format": "by_pair"},
    ).structured_content == {"jacobian": {}, "messages": []}



from __future__ import annotations

from collections.abc import Callable
from typing import cast

import numpy as np
import pytest

from pycycle_mcp.session_manager import session_manager
from pycycle_mcp.tools import create_model, derivatives, execution, sweep, variables
from pycycle_mcp.types import CycleProblem

from .conftest import DummyModel, DummyProblem


class _SetValFailsProblem(DummyProblem):
    def set_val(self, name: str | None = None, value: object | None = None, **kwargs: object) -> None:
        if name == "bad":
            raise RuntimeError("cannot set variable")
        super().set_val(name=name, value=value, **kwargs)


class _RunFailsProblem(DummyProblem):
    def set_solver_print(self, level: int = 0) -> None:
        del level
        raise RuntimeError("solver setup failed")


class _SweepSetValFailsProblem(DummyProblem):
    def set_val(self, name: str | None = None, value: object | None = None, **kwargs: object) -> None:
        del name, value, kwargs
        raise RuntimeError("sweep set failed")


def test_create_cycle_model_custom_requires_module_path() -> None:
    response = create_model.create_cycle_model({"cycle_type": "custom", "mode": "design"})

    assert response["error"]["type"] == "ValueError"


def test_create_cycle_model_with_builtin_type_uses_resolver(monkeypatch: pytest.MonkeyPatch) -> None:
    def fake_builder() -> DummyModel:
        return DummyModel(inputs=[("Mach", {"units": ""})], outputs=[("Fn", {"units": "lbf"})])

    def fake_build_problem(
        builder: Callable[[], DummyModel],
        mode: str,
        options: dict[str, object],
    ) -> tuple[DummyProblem, str]:
        del mode, options
        problem = DummyProblem()
        problem.model = builder()
        return problem, "builtin-dummy"

    monkeypatch.setattr(create_model, "_resolve_builtin_cycle", lambda cycle_type: (cycle_type, fake_builder))
    monkeypatch.setattr(create_model, "_build_problem", fake_build_problem)

    response = create_model.create_cycle_model({"cycle_type": "turbofan", "mode": "design", "options": {}})

    assert response["model_name"] == "builtin-dummy"
    assert "session_id" in response


def test_create_cycle_model_returns_structured_error_on_build_failure(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(create_model, "load_callable", lambda _: lambda: DummyModel())
    monkeypatch.setattr(
        create_model,
        "_build_problem",
        lambda builder, mode, options: (_ for _ in ()).throw(RuntimeError("build failed")),
    )

    response = create_model.create_cycle_model(
        {
            "cycle_type": "custom",
            "cycle_module_path": "unused.path",
            "mode": "design",
            "options": {},
        }
    )

    assert response["error"]["type"] == "RuntimeError"


def test_close_cycle_model_validation_and_error_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    assert create_model.close_cycle_model({})["error"]["type"] == "ValidationError"

    monkeypatch.setattr(create_model.session_manager, "close", lambda _: (_ for _ in ()).throw(RuntimeError("x")))
    response = create_model.close_cycle_model({"session_id": "abc"})
    assert response["error"]["type"] == "RuntimeError"


def test_get_cycle_summary_validation_and_missing_session_paths() -> None:
    assert create_model.get_cycle_summary({})["error"]["type"] == "ValidationError"

    response = create_model.get_cycle_summary({"session_id": "missing-id"})
    assert response["error"]["type"] == "KeyError"


def test_list_variables_validation_and_missing_session_paths() -> None:
    assert variables.list_variables({})["error"]["type"] == "ValidationError"

    response = variables.list_variables({"session_id": "missing-id"})
    assert response["error"]["type"] == "KeyError"


def test_list_variables_filters_and_limit_paths() -> None:
    problem = DummyProblem()
    problem.model.inputs = [
        ("keep_input", {"promoted_name": "keep_input", "units": "ft"}),
        ("skip_promoted_input", {"promoted_name": "different_name"}),
    ]
    problem.model.outputs = [
        ("keep_output", {"promoted_name": "keep_output", "units": "lbf"}),
        ("skip_promoted_output", {"promoted_name": "different_output"}),
        ("other_output", {"promoted_name": "other_output"}),
    ]
    session_id = session_manager.create_session(problem=cast(CycleProblem, problem), meta={})

    response = variables.list_variables(
        {
            "session_id": session_id,
            "kind": "both",
            "promoted_only": True,
            "name_filter": "keep",
            "max_variables": "1",
        }
    )
    assert len(response["variables"]) == 1
    assert response["variables"][0]["name"] in {"keep_input", "keep_output"}


def test_set_inputs_validation_and_error_paths() -> None:
    assert variables.set_inputs({})["error"]["type"] == "ValidationError"
    assert variables.set_inputs({"session_id": "x", "values": {}})["error"]["type"] == "ValidationError"
    assert variables.set_inputs({"session_id": "missing-id", "values": {"x": 1.0}})["error"]["type"] == "KeyError"

    problem = _SetValFailsProblem()
    session_id = session_manager.create_session(problem=cast(CycleProblem, problem), meta={})

    fail_response = variables.set_inputs({"session_id": session_id, "values": {"bad": 1.0}})
    assert fail_response["error"]["type"] == "RuntimeError"

    allow_response = variables.set_inputs(
        {"session_id": session_id, "values": {"bad": 1.0, "ok": 2.0}, "allow_missing": True}
    )
    assert allow_response["updated"] == ["ok"]
    assert allow_response["skipped"][0]["name"] == "bad"


def test_get_outputs_validation_and_lookup_paths() -> None:
    assert variables.get_outputs({})["error"]["type"] == "ValidationError"
    assert variables.get_outputs({"session_id": "x", "names": []})["error"]["type"] == "ValidationError"
    assert variables.get_outputs({"session_id": "missing-id", "names": ["Fn"]})["error"]["type"] == "KeyError"

    problem = DummyProblem()
    problem.values["Fn"] = 12.3
    session_id = session_manager.create_session(problem=cast(CycleProblem, problem), meta={})

    response = variables.get_outputs({"session_id": session_id, "names": ["missing"]})
    assert response["error"]["type"] == "LookupError"


def test_run_cycle_error_paths_and_driver_branch() -> None:
    assert execution.run_cycle({})["error"]["type"] == "ValidationError"
    assert execution.run_cycle({"session_id": "missing-id"})["error"]["type"] == "KeyError"

    run_fail_session = session_manager.create_session(problem=cast(CycleProblem, _RunFailsProblem()), meta={})
    run_fail_response = execution.run_cycle({"session_id": run_fail_session})
    assert run_fail_response["error"]["type"] == "RuntimeError"

    driver_problem = DummyProblem()
    driver_problem.values["Fn"] = 44.0
    driver_session = session_manager.create_session(problem=cast(CycleProblem, driver_problem), meta={})
    driver_response = execution.run_cycle(
        {
            "session_id": driver_session,
            "outputs_of_interest": ["Fn", "missing_output"],
            "use_driver": True,
        }
    )
    assert driver_response["success"] is True
    assert driver_response["messages"][0] == "Ran driver"
    assert driver_response["outputs"]["missing_output"] is None


def test_compute_totals_validation_dense_and_missing_session_paths() -> None:
    assert derivatives.compute_totals({})["error"]["type"] == "ValidationError"
    assert derivatives.compute_totals({"session_id": "x", "of": [], "wrt": []})["error"]["type"] == "ValidationError"

    problem = DummyProblem()
    session_id = session_manager.create_session(problem=cast(CycleProblem, problem), meta={})
    dense_response = derivatives.compute_totals(
        {"session_id": session_id, "of": ["Fn"], "wrt": ["Mach"], "return_format": "dense"}
    )
    assert dense_response["jacobian"] == {"of": ["Fn"], "wrt": ["Mach"], "data": [[[[1.0]]]]}

    missing_response = derivatives.compute_totals({"session_id": "missing", "of": ["Fn"], "wrt": ["Mach"]})
    assert missing_response["error"]["type"] == "KeyError"


def test_sweep_inputs_validation_and_failure_paths() -> None:
    assert sweep.sweep_inputs({})["error"]["type"] == "ValidationError"
    assert sweep.sweep_inputs({"session_id": "x", "sweep": []})["error"]["type"] == "ValidationError"
    assert (
        sweep.sweep_inputs({"session_id": "missing-id", "sweep": [{"name": "x", "values": [1]}]})["error"]["type"]
        == "KeyError"
    )

    failing_problem = _SweepSetValFailsProblem()
    session_id = session_manager.create_session(problem=cast(CycleProblem, failing_problem), meta={})

    skip_response = sweep.sweep_inputs(
        {
            "session_id": session_id,
            "sweep": [{"name": "Mach", "values": [0.7]}],
            "skip_on_failure": True,
        }
    )
    assert skip_response["results"][0]["success"] is False

    fail_fast_response = sweep.sweep_inputs(
        {
            "session_id": session_id,
            "sweep": [{"name": "Mach", "values": [0.7]}],
            "skip_on_failure": False,
        }
    )
    assert fail_fast_response["error"]["type"] == "RuntimeError"


@pytest.mark.parametrize(
    "value,expected",
    [
        (42, 42),
        (3.5, 3.5),
        (np.array([4.2]), 4.2),
        (np.array([1.0, 2.0]), [1.0, 2.0]),
        (np.int64(7), 7),
        (np.float64(8.5), 8.5),
    ],
)
def test_to_serializable_passthrough_scalars(value: object, expected: object) -> None:
    assert execution._to_serializable(value) == expected



from __future__ import annotations

from typing import cast

from pycycle_mcp.session_manager import session_manager
from pycycle_mcp.tools import execution, variables
from pycycle_mcp.types import CycleProblem

from .conftest import DummyProblem


def setup_dummy_session() -> str:
    problem = DummyProblem()
    problem.model.inputs = [
        (
            "Mach",
            {"promoted_name": "Mach", "units": "", "desc": "Mach number", "val": 0.0},
        )
    ]
    problem.model.outputs = [("Fn", {"promoted_name": "Fn", "units": "lbf", "desc": "Thrust", "val": 0.0})]
    return session_manager.create_session(problem=cast(CycleProblem, problem), meta={"mode": "design", "options": {}})


def test_set_and_get_inputs() -> None:
    session_id = setup_dummy_session()

    set_response = variables.set_inputs({"session_id": session_id, "values": {"Mach": 0.75}})
    assert set_response["updated"] == ["Mach"]

    list_response = variables.list_variables({"session_id": session_id, "kind": "inputs"})
    variables_payload = cast(list[dict[str, object]], list_response["variables"])
    assert variables_payload[0]["name"] == "Mach"


def test_run_cycle_and_get_outputs() -> None:
    session_id = setup_dummy_session()

    run_response = execution.run_cycle({"session_id": session_id, "outputs_of_interest": ["Fn"]})
    assert run_response["success"] is True

    # populate output for retrieval
    problem, _ = session_manager.get(session_id)
    assert isinstance(problem, DummyProblem)
    problem.values["Fn"] = 123.4

    output_response = variables.get_outputs({"session_id": session_id, "names": ["Fn"]})
    values = cast(dict[str, object], output_response["values"])
    assert values["Fn"] == 123.4


def test_get_outputs_missing_allowed() -> None:
    session_id = setup_dummy_session()
    response = variables.get_outputs({"session_id": session_id, "names": ["Missing"], "allow_missing": True})
    assert response["missing"] == ["Missing"]



import sys

import openmdao.api as om

import pycycle.api as pyc


class Turbojet(pyc.Cycle):

    def setup(self):

        USE_TABULAR = True

        if USE_TABULAR: 
            self.options['thermo_method'] = 'TABULAR'
            self.options['thermo_data'] = pyc.AIR_JETA_TAB_SPEC
            FUEL_TYPE = "FAR"
        else: 
            self.options['thermo_method'] = 'CEA'
            self.options['thermo_data'] = pyc.species_data.janaf
            FUEL_TYPE = "Jet-A(g)"

        design = self.options['design']

        # Add engine elements
        self.add_subsystem('fc', pyc.FlightConditions())
        self.add_subsystem('inlet', pyc.Inlet())
        self.add_subsystem('comp', pyc.Compressor(map_data=pyc.AXI5, map_extrap=True),
                                    promotes_inputs=['Nmech'])
        self.add_subsystem('burner', pyc.Combustor(fuel_type=FUEL_TYPE))
        self.add_subsystem('turb', pyc.Turbine(map_data=pyc.LPT2269),
                                    promotes_inputs=['Nmech'])
        self.add_subsystem('nozz', pyc.Nozzle(nozzType='CD', lossCoef='Cv'))
        self.add_subsystem('shaft', pyc.Shaft(num_ports=2),promotes_inputs=['Nmech'])
        self.add_subsystem('perf', pyc.Performance(num_nozzles=1, num_burners=1))

        # Connect flow stations
        self.pyc_connect_flow('fc.Fl_O', 'inlet.Fl_I', connect_w=False)
        self.pyc_connect_flow('inlet.Fl_O', 'comp.Fl_I')
        self.pyc_connect_flow('comp.Fl_O', 'burner.Fl_I')
        self.pyc_connect_flow('burner.Fl_O', 'turb.Fl_I')
        self.pyc_connect_flow('turb.Fl_O', 'nozz.Fl_I')

        # Make other non-flow connections
        # Connect turbomachinery elements to shaft
        self.connect('comp.trq', 'shaft.trq_0')
        self.connect('turb.trq', 'shaft.trq_1')

        # Connnect nozzle exhaust to freestream static conditions
        self.connect('fc.Fl_O:stat:P', 'nozz.Ps_exhaust')

        # Connect outputs to perfomance element
        self.connect('inlet.Fl_O:tot:P', 'perf.Pt2')
        self.connect('comp.Fl_O:tot:P', 'perf.Pt3')
        self.connect('burner.Wfuel', 'perf.Wfuel_0')
        self.connect('inlet.F_ram', 'perf.ram_drag')
        self.connect('nozz.Fg', 'perf.Fg_0')

        # Add balances for design and off-design
        balance = self.add_subsystem('balance', om.BalanceComp())
        if design:

            balance.add_balance('W', units='lbm/s', eq_units='lbf', rhs_name='Fn_target')
            self.connect('balance.W', 'inlet.Fl_I:stat:W')
            self.connect('perf.Fn', 'balance.lhs:W')

            balance.add_balance('FAR', eq_units='degR', lower=1e-4, val=.017, rhs_name='T4_target')
            self.connect('balance.FAR', 'burner.Fl_I:FAR')
            self.connect('burner.Fl_O:tot:T', 'balance.lhs:FAR')

            balance.add_balance('turb_PR', val=1.5, lower=1.001, upper=8, eq_units='hp', rhs_val=0.)
            self.connect('balance.turb_PR', 'turb.PR')
            self.connect('shaft.pwr_net', 'balance.lhs:turb_PR')

        else:

            balance.add_balance('FAR', eq_units='lbf', lower=1e-4, val=.3, rhs_name='Fn_target')
            self.connect('balance.FAR', 'burner.Fl_I:FAR')
            self.connect('perf.Fn', 'balance.lhs:FAR')

            balance.add_balance('Nmech', val=1.5, units='rpm', lower=500., eq_units='hp', rhs_val=0.)
            self.connect('balance.Nmech', 'Nmech')
            self.connect('shaft.pwr_net', 'balance.lhs:Nmech')

            balance.add_balance('W', val=168.0, units='lbm/s', eq_units='inch**2')
            self.connect('balance.W', 'inlet.Fl_I:stat:W')
            self.connect('nozz.Throat:stat:area', 'balance.lhs:W')

       
        newton = self.nonlinear_solver = om.NewtonSolver()
        newton.options['atol'] = 1e-6
        newton.options['rtol'] = 1e-6
        newton.options['iprint'] = 2
        newton.options['maxiter'] = 15
        newton.options['solve_subsystems'] = True
        newton.options['max_sub_solves'] = 100
        newton.options['reraise_child_analysiserror'] = False
        
        self.linear_solver = om.DirectSolver()

        super().setup()

def viewer(prob, pt, file=sys.stdout):
    """
    print a report of all the relevant cycle properties
    """

    summary_data = (prob[pt+'.fc.Fl_O:stat:MN'], prob[pt+'.fc.alt'], prob[pt+'.inlet.Fl_O:stat:W'], 
                    prob[pt+'.perf.Fn'], prob[pt+'.perf.Fg'], prob[pt+'.inlet.F_ram'],
                    prob[pt+'.perf.OPR'], prob[pt+'.perf.TSFC'])

    print(file=file, flush=True)
    print(file=file, flush=True)
    print(file=file, flush=True)
    print("----------------------------------------------------------------------------", file=file, flush=True)
    print("                              POINT:", pt, file=file, flush=True)
    print("----------------------------------------------------------------------------", file=file, flush=True)
    print("                       PERFORMANCE CHARACTERISTICS", file=file, flush=True)
    print("    Mach      Alt       W      Fn      Fg    Fram     OPR     TSFC  ", file=file, flush=True)
    print(" %7.5f  %7.1f %7.3f %7.1f %7.1f %7.1f %7.3f  %7.5f" %summary_data, file=file, flush=True)


    fs_names = ['fc.Fl_O', 'inlet.Fl_O', 'comp.Fl_O', 'burner.Fl_O',
                'turb.Fl_O', 'nozz.Fl_O']
    fs_full_names = [f'{pt}.{fs}' for fs in fs_names]
    pyc.print_flow_station(prob, fs_full_names, file=file)

    comp_names = ['comp']
    comp_full_names = [f'{pt}.{c}' for c in comp_names]
    pyc.print_compressor(prob, comp_full_names, file=file)

    pyc.print_burner(prob, [f'{pt}.burner'])

    turb_names = ['turb']
    turb_full_names = [f'{pt}.{t}' for t in turb_names]
    pyc.print_turbine(prob, turb_full_names, file=file)

    noz_names = ['nozz']
    noz_full_names = [f'{pt}.{n}' for n in noz_names]
    pyc.print_nozzle(prob, noz_full_names, file=file)

    shaft_names = ['shaft']
    shaft_full_names = [f'{pt}.{s}' for s in shaft_names]
    pyc.print_shaft(prob, shaft_full_names, file=file)

def map_plots(prob, pt):
    comp_names = ['comp']
    comp_full_names = [f'{pt}.{c}' for c in comp_names]
    pyc.plot_compressor_maps(prob, comp_full_names)

    turb_names = ['turb']
    turb_full_names = [f'{pt}.{c}' for c in turb_names]
    pyc.plot_turbine_maps(prob, turb_full_names)



class MPTurbojet(pyc.MPCycle):

    def setup(self):

        # Create design instance of model
        self.pyc_add_pnt('DESIGN', Turbojet())

        self.set_input_defaults('DESIGN.Nmech', 8070.0, units='rpm')
        self.set_input_defaults('DESIGN.inlet.MN', 0.60)
        self.set_input_defaults('DESIGN.comp.MN', 0.020)#.2
        self.set_input_defaults('DESIGN.burner.MN', 0.020)#.2
        self.set_input_defaults('DESIGN.turb.MN', 0.4)

        self.pyc_add_cycle_param('burner.dPqP', 0.03)
        self.pyc_add_cycle_param('nozz.Cv', 0.99)

        
        # define the off-design conditions we want to run
        self.od_pts = ['OD0', 'OD1']
        self.od_MNs = [0.000001, 0.2]
        self.od_alts = [0.0, 5000]
        self.od_Fns =[11000.0, 8000.0]

        for i,pt in enumerate(self.od_pts):
            self.pyc_add_pnt(pt, Turbojet(design=False))

            self.set_input_defaults(pt+'.fc.MN', val=self.od_MNs[i])
            self.set_input_defaults(pt+'.fc.alt', self.od_alts[i], units='ft')
            self.set_input_defaults(pt+'.balance.Fn_target', self.od_Fns[i], units='lbf')  

        self.pyc_use_default_des_od_conns()

        self.pyc_connect_des_od('nozz.Throat:stat:area', 'balance.rhs:W')

        super().setup()

if __name__ == "__main__":

    import time

    prob = om.Problem()


    mp_turbojet = prob.model = MPTurbojet()

    # prob.model.set_order(['DESIGN', 'OD0', 'OD1', 'test')

    prob.setup(check=False)

    #Define the design point
    prob.set_val('DESIGN.fc.alt', 0, units='ft')
    prob.set_val('DESIGN.fc.MN', 0.000001)
    prob.set_val('DESIGN.balance.Fn_target', 11800.0, units='lbf')
    prob.set_val('DESIGN.balance.T4_target', 2370.0, units='degR') 
    prob.set_val('DESIGN.comp.PR', 13.5) 
    prob.set_val('DESIGN.comp.eff', 0.83)
    prob.set_val('DESIGN.turb.eff', 0.86)

    # Set initial guesses for balances
    prob['DESIGN.balance.FAR'] = 0.0175506829934
    prob['DESIGN.balance.W'] = 168.453135137
    prob['DESIGN.balance.turb_PR'] = 4.46138725662
    prob['DESIGN.fc.balance.Pt'] = 14.6955113159
    prob['DESIGN.fc.balance.Tt'] = 518.665288153

    for i,pt in enumerate(mp_turbojet.od_pts):

        # initial guesses
        prob[pt+'.balance.W'] = 166.073
        prob[pt+'.balance.FAR'] = 0.01680
        prob[pt+'.balance.Nmech'] = 8197.38
        prob[pt+'.fc.balance.Pt'] = 15.703
        prob[pt+'.fc.balance.Tt'] = 558.31
        prob[pt+'.turb.PR'] = 4.6690

    st = time.time()

    prob.set_solver_print(level=-1)
    prob.set_solver_print(level=2, depth=1)

    # prob.model.OD1.nonlinear_solver.options['maxiter'] = 1

    prob.run_model()

    # prob.model.OD1.list_outputs(residuals=True)
    # exit()

    for pt in ['DESIGN']+mp_turbojet.od_pts:
        viewer(prob, pt)

    map_plots(prob,'DESIGN')

    print()
    print("time", time.time() - st)


"""Entry point for the pyCycle MCP FastMCP server."""

from __future__ import annotations

import argparse
from collections.abc import Sequence
from typing import Any, Literal, cast

from pycycle_mcp.fastmcp_server import build_server

TransportName = Literal["stdio", "http", "sse", "streamable-http"]


def build_parser() -> argparse.ArgumentParser:
    """Create the argument parser for the FastMCP server CLI."""
    parser = argparse.ArgumentParser(description="pyCycle MCP server")
    parser.add_argument(
        "--transport",
        choices=("stdio", "http", "sse", "streamable-http"),
        default="stdio",
        help=(
            "Transport for serving MCP (stdio for local/editor integrations, "
            "HTTP-compatible modes for network serving)."
        ),
    )
    parser.add_argument(
        "--host",
        default="0.0.0.0",
        help="Host to bind for HTTP-compatible transports (default: 0.0.0.0)",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=8000,
        help="Port to bind for HTTP-compatible transports (default: 8000)",
    )
    parser.add_argument(
        "--path",
        help=("Optional path to mount the HTTP endpoint (defaults to FastMCP's protocol-specific path)."),
    )
    parser.add_argument(
        "--log-level",
        default="INFO",
        help="Logging level for server startup messages.",
    )
    return parser


def _normalize_transport(transport: TransportName) -> TransportName:
    """Normalize legacy HTTP aliasing used by existing integrations."""
    if transport == "http":
        return "streamable-http"
    return transport


def main(argv: Sequence[str] | None = None) -> int:
    """Register tools and start the FastMCP server."""
    parser = build_parser()
    args = parser.parse_args(argv)

    app = build_server()
    transport = _normalize_transport(cast(TransportName, args.transport))

    transport_kwargs: dict[str, Any] = {
        "show_banner": False,
        "log_level": args.log_level,
    }
    if transport in {"sse", "streamable-http"}:
        transport_kwargs["host"] = args.host
        transport_kwargs["port"] = args.port
        if args.path is not None:
            transport_kwargs["path"] = args.path

    app.run(transport=transport, **transport_kwargs)
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""Sweep tools for parametric analyses."""

from __future__ import annotations

from ..errors import error_response, to_error
from ..session_manager import session_manager
from ..utils import error_on_missing_session, ordered_cartesian_product
from .execution import run_cycle


def sweep_inputs(payload: dict[str, object]) -> dict[str, object]:
    """Perform a parametric sweep over input variables."""

    session_id = payload.get("session_id")
    sweep_spec_raw = payload.get("sweep") or []
    sweep_spec = sweep_spec_raw if isinstance(sweep_spec_raw, list) else []
    outputs_raw = payload.get("outputs_of_interest") or []
    outputs_of_interest: list[str] = outputs_raw if isinstance(outputs_raw, list) else []
    use_driver = bool(payload.get("use_driver", False))
    skip_on_failure = bool(payload.get("skip_on_failure", True))

    if not session_id:
        return error_response("ValidationError", "session_id is required")
    if not sweep_spec:
        return error_response("ValidationError", "sweep must include at least one variable")

    try:
        problem, _ = session_manager.get(str(session_id))
        variables = [entry.get("name") for entry in sweep_spec if isinstance(entry, dict)]
        value_sets = [
            entry.get("values", []) if isinstance(entry.get("values"), list) else []
            for entry in sweep_spec
            if isinstance(entry, dict)
        ]

        results: list[dict[str, object]] = []
        for value_combo in ordered_cartesian_product(value_sets):
            input_values = {
                name: value for name, value in zip(variables, value_combo, strict=False) if isinstance(name, str)
            }
            try:
                for var_name, var_value in input_values.items():
                    problem.set_val(var_name, var_value)
                run_result = run_cycle(
                    {
                        "session_id": session_id,
                        "outputs_of_interest": outputs_of_interest,
                        "use_driver": use_driver,
                    }
                )
                success = not run_result.get("error") and bool(run_result.get("success"))
                results.append(
                    {
                        "inputs": input_values,
                        "success": success,
                        "outputs": run_result.get("outputs", {}),
                        "error_message": (None if success else str(run_result.get("error"))),
                    }
                )
            except Exception as exc:
                if skip_on_failure:
                    results.append(
                        {
                            "inputs": input_values,
                            "success": False,
                            "outputs": {},
                            "error_message": str(exc),
                        }
                    )
                    continue
                return to_error(exc)

        return {"results": results}
    except KeyError as exc:
        return error_on_missing_session(str(session_id), exc)
    except Exception as exc:  # pragma: no cover
        return to_error(exc)



"""Session management for pyCycle MCP server."""

from __future__ import annotations

import uuid
from dataclasses import dataclass, field

from .types import CycleProblem


@dataclass
class SessionRecord:
    """Container for an OpenMDAO Problem and metadata."""

    problem: CycleProblem
    meta: dict[str, object] = field(default_factory=dict)


class SessionManager:
    """Manage pyCycle/OpenMDAO Problem sessions keyed by UUID."""

    def __init__(self) -> None:
        self._sessions: dict[str, SessionRecord] = {}

    def create_session(self, problem: CycleProblem, meta: dict[str, object] | None = None) -> str:
        """Register a new session and return its identifier."""

        session_id = str(uuid.uuid4())
        self._sessions[session_id] = SessionRecord(problem=problem, meta=meta or {})
        return session_id

    def get(self, session_id: str) -> tuple[CycleProblem, dict[str, object]]:
        """Retrieve the problem and metadata for a session."""

        record = self._sessions.get(session_id)
        if record is None:
            raise KeyError(f"Unknown session_id: {session_id}")
        return record.problem, record.meta

    def close(self, session_id: str) -> None:
        """Remove a session."""

        if session_id in self._sessions:
            del self._sessions[session_id]


session_manager = SessionManager()
"""Module-level session manager for convenience."""



"""Execution tools for running cycle models."""

from __future__ import annotations

import logging

import numpy as np

from ..errors import error_response, to_error
from ..session_manager import session_manager
from ..utils import error_on_missing_session

LOGGER = logging.getLogger(__name__)

DEFAULT_OUTPUTS = [
    "perf.Fn",
    "perf.TSFC",
    "perf.OPR",
    "fc.Fl_O:stat:MN",
    "fc.alt",
    "splitter.BPR",
]


def _to_serializable(val: object) -> object:
    """Convert numpy types to native Python for JSON serialization."""
    if isinstance(val, np.ndarray):
        flat = val.flatten()
        if flat.size == 1:
            return float(flat[0])
        return [float(v) for v in flat]
    if isinstance(val, (np.floating, np.complexfloating)):
        return float(val)
    if isinstance(val, np.integer):
        return int(val)
    return val


def run_cycle(payload: dict[str, object]) -> dict[str, object]:
    """Run the cycle model and return selected outputs."""

    session_id = payload.get("session_id")
    outputs_of_interest: list[str] = payload.get("outputs_of_interest") or DEFAULT_OUTPUTS  # type: ignore[assignment]
    use_driver = bool(payload.get("use_driver", False))

    if not session_id:
        return error_response("ValidationError", "session_id is required")

    try:
        problem, _ = session_manager.get(str(session_id))
        messages: list[str] = []
        try:
            problem.set_solver_print(level=-1)
            if use_driver:
                messages.append("Ran driver")
                problem.run_driver()
            else:
                messages.append("Ran model")
                problem.run_model()
        except Exception as exc:
            LOGGER.error("Run failed: %s", exc)
            return to_error(exc)

        outputs: dict[str, object | None] = {}
        for name in outputs_of_interest:
            try:
                outputs[name] = _to_serializable(problem.get_val(name))
            except Exception as exc:
                outputs[name] = None
                messages.append(f"Missing output {name}: {exc}")

        return {
            "success": True,
            "outputs": outputs,
            "messages": messages,
        }
    except KeyError as exc:
        return error_on_missing_session(str(session_id), exc)
    except Exception as exc:  # pragma: no cover
        return to_error(exc)
